#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This is the default collation function (IcuNoCase).
"""

__author__ = "Alberto Pettarin"
__copyright__ = "Copyright 2012-2015, Alberto Pettarin (www.albertopettarin.it)"
__license__ = "MIT"
__version__ = "3.1.0"
__email__ = "alberto@albertopettarin.it"
__status__ = "Production"

def collate_function(string1, string2):
    """
    Implement default IcuNoCase collation,
    by simply lowercasing the UTF-8 encoded versions
    of the two strings.

    :param string1: first string
    :type  string1: unicode
    :param string2: second string
    :type  string2: unicode
    :rtype: int
    """
    b1 = string1.encode("utf-8").lower()
    b2 = string2.encode("utf-8").lower()
    if b1 == b2:
        return 0
    if b1 < b2:
        return -1
    return 1






#!/usr/bin/env python
# coding=utf-8

"""
This file contains a collection of miscellaneous utility functions.
"""

from __future__ import absolute_import
from __future__ import print_function
from io import open
import imp
import os
import shutil
import tempfile
import uuid
import zipfile

__author__ = "Alberto Pettarin"
__copyright__ = "Copyright 2012-2015, Alberto Pettarin (www.albertopettarin.it)"
__license__ = "MIT"
__version__ = "3.1.0"
__email__ = "alberto@albertopettarin.it"
__status__ = "Production"

def print_debug(msg, do_print=True):
    if do_print:
        print(u"[DEBU] %s" % msg)

def print_error(msg):
    print(u"[ERRO] %s" % msg)

def print_info(msg):
    print(u"[INFO] %s" % msg)

def get_uuid():
    return str(uuid.uuid4()).replace("-", "")

def load_input_parser(parser_file_path):
    parser = None
    if os.path.exists(parser_file_path):
        try:
            # load source file
            parser = imp.load_source("", parser_file_path)
            try:
                # try calling parse function
                parser.parse(None, None)
            except:
                print_error("Error trying to call the parse() function. Does file '%s' contain a parse() function?" % parser_file_path)
        except:
            print_error("Error trying to load parser from file '%s'" % parser_file_path)
    else:
        print_error("File '%s' does not exist" % parser_file_path)
    return parser

def create_temp_file():
    tmp_handler, tmp_path = tempfile.mkstemp()
    return (tmp_handler, tmp_path)

def create_temp_directory():
    return tempfile.mkdtemp()

def copy_file(origin, destination):
    try:
        shutil.copy(origin, destination)
    except:
        pass

def rename_file(origin, destination):
    try:
        os.rename(origin, destination)
    except:
        pass

def delete_file(handler, path):
    """
    Safely delete file.

    :param handler: the file handler (as returned by tempfile)
    :type  handler: obj
    :param path: the file path
    :type  path: string (path)
    """
    if handler is not None:
        try:
            os.close(handler)
        except:
            pass
    if path is not None:
        try:
            os.remove(path)
        except:
            pass

def delete_directory(path):
    """
    Safely delete a directory.

    :param path: the file path
    :type  path: string (path)
    """
    if path is not None:
        try:
            shutil.rmtree(path)
        except:
            pass






#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This is a collation function (IcuNoCase) for German.
"""

__author__ = "Alberto Pettarin"
__copyright__ = "Copyright 2012-2015, Alberto Pettarin (www.albertopettarin.it)"
__license__ = "MIT"
__version__ = "3.1.0"
__email__ = "alberto@albertopettarin.it"
__status__ = "Production"

REPLACEMENTS = [
    [u"ä", u"a"],
    [u"ö", u"o"],
    [u"ü", u"u"],
    [u"ß", u"ss"]
]

def collate_function(string1, string2):
    """
    Implement IcuNoCase collation for German.
    (I do not remember where the procedure comes from.)

    :param string1: first string
    :type  string1: unicode
    :param string2: second string
    :type  string2: unicode
    :rtype: int
    """
    b1 = string1.lower()
    b2 = string2.lower()
    c1 = b1
    c2 = b2
    for repl in REPLACEMENTS:
        b1 = b1.replace(repl[0], repl[1])
        b2 = b2.replace(repl[0], repl[1])
    if b1.encode("utf-16") == b2.encode("utf-16"):
        if c1.encode("utf-16") == c2.encode("utf-16"):
            return 0
        if c1.encode("utf-16") < c2.encode("utf-16"):
            return -1
    else:
        if b1.encode("utf-16") < b2.encode("utf-16"):
            return -1
    return 1






#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This is the default prefix function for grouping headwords.
"""

__author__ = "Alberto Pettarin"
__copyright__ = "Copyright 2012-2015, Alberto Pettarin (www.albertopettarin.it)"
__license__ = "MIT"
__version__ = "3.1.0"
__email__ = "alberto@albertopettarin.it"
__status__ = "Production"

def get_prefix(headword, length):
    """
    Return the prefix for the given headword,
    of length length.

    :param headword: the headword string
    :type  headword: unicode
    :param length: prefix length
    :type  length: int
    :rtype: unicode
    """
    if headword is None:
        return None
    lowercased = headword.lower()
    if ord(lowercased[0]) < 97:
        return u"SPECIAL"
    if len(lowercased) < length:
        return lowercased
    return lowercased[0:length]





#!/usr/bin/env python
# coding=utf-8

"""
Penelope is a multi-tool for creating, editing and converting dictionaries, especially for eReader devices.

This is the main Penelope script, intended to be run from command line.
"""

from __future__ import absolute_import
import argparse
import sys

from penelope.__main__ import main
from penelope.command_line import COMMAND_LINE_PARAMETERS
from penelope.command_line import DESCRIPTION
from penelope.command_line import EPILOG
from penelope.command_line import INPUT_FORMATS
from penelope.command_line import OUTPUT_FORMATS
from penelope.command_line import REQUIRED_PARAMETERS
from penelope.command_line import USAGE
from penelope.command_line import check_arguments
from penelope.command_line import set_default_values
from penelope.dictionary import read_dictionary
from penelope.dictionary import write_dictionary
from penelope.utilities import get_uuid
from penelope.utilities import load_input_parser
from penelope.utilities import print_debug
from penelope.utilities import print_error
from penelope.utilities import print_info

__author__ = "Alberto Pettarin"
__copyright__ = "Copyright 2012-2015, Alberto Pettarin (www.albertopettarin.it)"
__license__ = "MIT"
__version__ = "3.1.0"
__email__ = "alberto@albertopettarin.it"
__status__ = "Production"






#!/usr/bin/env python
# coding=utf-8

"""
This is an example of an input parser, acting on a real-world dictionary,
the Webster 1913.
"""

__author__ = "Alberto Pettarin"
__copyright__ = "Copyright 2012-2015, Alberto Pettarin (www.albertopettarin.it)"
__license__ = "MIT"
__version__ = "3.1.0"
__email__ = "alberto@albertopettarin.it"
__status__ = "Production"

def parse(dictionary, arguments):
    """
    Given the input dictionary and arguments,
    return a (possibly, modified) dictionary.

    The returned dictionary might be the same (input) instance,
    or a new one.

    :param dictionary: the (raw) input dictionary
    :type  dictionary: Dictionary
    :param arguments: the command line arguments
    :type  arguments: Namespace from argparse
    :rtype: Dictionary
    """

    def clean_definition(entry):
        """
        Remove any prefix of type "<k>...</k>"
        from the entry.definition string.
        """
        pos = entry.definition.find("</k>")
        if pos > -1:
            entry.definition = entry.definition[pos+len("</k>"):].strip()

    def custom_merge_function(headword, definitions):
        """
        Merge definitions for the same headword in a custom way:
        1 def   => <b>word<b>
                    <f>&nbsp;&ns;&nbsp;</f>definition
        2+ defs => <b>word<b>
                    <f>&nbsp;&ns;&nbsp;</f><b>(1)&nbsp;definition
                    <f>&nbsp;&ns;&nbsp;</f><b>(2)&nbsp;definition...
        """
        if len(definitions) > 1:
            clean = u"<b>%s<b>" % (headword)
            for i in range(len(definitions)):
                clean += "<f>&nbsp;&ns;&nbsp;</f><b>(%d)</b>&nbsp;%s" % (i+1, definitions[i])
            return clean
        if len(definitions) == 1:
            return u"<b>%s<b><f>&nbsp;&ns;&nbsp;</f>%s" % (headword, definitions[0])
        return u""

    if dictionary is None:
        return None
    for entry in dictionary.entries:
        clean_definition(entry)
    dictionary.merge_definitions(merge_function=custom_merge_function)
    return dictionary






#!/usr/bin/env python
# coding=utf-8

"""
Penelope is a multi-tool for creating, editing and converting dictionaries, especially for eReader devices.

This is the main Penelope script, intended to be run from command line.
"""

from __future__ import absolute_import
import argparse
import sys

from penelope.command_line import COMMAND_LINE_PARAMETERS
from penelope.command_line import DESCRIPTION
from penelope.command_line import EPILOG
from penelope.command_line import INPUT_FORMATS
from penelope.command_line import OUTPUT_FORMATS
from penelope.command_line import REQUIRED_PARAMETERS
from penelope.command_line import USAGE
from penelope.command_line import check_arguments
from penelope.command_line import set_default_values
from penelope.dictionary import read_dictionary
from penelope.dictionary import write_dictionary
from penelope.utilities import get_uuid
from penelope.utilities import load_input_parser
from penelope.utilities import print_debug
from penelope.utilities import print_error
from penelope.utilities import print_info

__author__ = "Alberto Pettarin"
__copyright__ = "Copyright 2012-2015, Alberto Pettarin (www.albertopettarin.it)"
__license__ = "MIT"
__version__ = "3.1.0"
__email__ = "alberto@albertopettarin.it"
__status__ = "Production"

def main():
    parser = argparse.ArgumentParser(
        usage=USAGE,
        description=DESCRIPTION,
        epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    for param in COMMAND_LINE_PARAMETERS:
        if param["short"] is None:
            parser.add_argument(
                param["long"],
                help=param["help"],
                action=param["action"],
                default=argparse.SUPPRESS
            )
        else:
            parser.add_argument(
                param["short"],
                param["long"],
                help=param["help"],
                action=param["action"],
                default=argparse.SUPPRESS
            )
    arguments = parser.parse_args()

    # no arguments: show help and exit
    if len(sys.argv) < 2:
        parser.print_help()
        sys.exit(0)

    # print version and exit
    if "version" in arguments:
        print_info("Penelope v%s" % (__version__))
        sys.exit(0)

    # check we have all the required arguments
    # if not, it will sys.exit() with some error code
    check_arguments(arguments)

    # set default values
    set_default_values(arguments)
    print_debug(u"Running with the command line arguments:\n%s" % (str(arguments)), arguments.debug)

    # read raw dictionary
    print_info(u"Reading input file(s)...")
    dictionary = read_dictionary(arguments)
    if dictionary is None:
        print_error("Unable to read the input file(s)")
        sys.exit(8)
    print_info(u"Reading input file(s)... done")

    # apply custom input parser, if specified
    if arguments.input_parser is not None:
        input_parser = load_input_parser(arguments.input_parser)
        if input_parser is not None:
            print_info(u"Applying the specified input parser...")
            dictionary = input_parser.parse(dictionary, arguments)
            print_info(u"Applying the specified input parser... done")

    # sort dictionary before, if requested
    if arguments.sort_before:
        print_info(u"Sorting before...")
        dictionary.sort(
            arguments.sort_by_headword,
            arguments.sort_by_definition,
            arguments.sort_reverse,
            arguments.sort_ignore_case
        )
        print_info(u"Sorting before... done")

    # merge definitions, if requested
    if arguments.merge_definitions:
        print_info(u"Merging...")
        dictionary.merge_definitions(merge_separator=arguments.merge_separator)
        print_info(u"Merging... done")

    # flatten synonyms, if requested
    if arguments.flatten_synonyms:
        print_info(u"Flattening synonyms...")
        dictionary.flatten_synonyms()
        print_info(u"Flattening synonyms... done")

    # sort dictionary after, if requested
    if arguments.sort_after:
        print_info(u"Sorting after...")
        dictionary.sort(
            arguments.sort_by_headword,
            arguments.sort_by_definition,
            arguments.sort_reverse,
            arguments.sort_ignore_case
        )
        print_info(u"Sorting after... done")

    # output dictionary
    print_info(u"Writing output file(s)...")
    output_paths = write_dictionary(dictionary, arguments)
    if output_paths is None:
        print_error("Unable to write the output file(s)")
        sys.exit(16)
    print_info(u"Writing output file(s)... done")
    print_info(u"The following file(s) have been created:")
    for op in output_paths:
        print_info(u"  %s" % op)

    sys.exit(0)


if __name__ == "__main__":
    main()






#!/usr/bin/env python
# coding=utf-8

"""
Write EPUB dictionaries.

NOTE: this file needs refactoring.
"""

from __future__ import absolute_import
from io import open
import os
import zipfile

from penelope.dictionary_ebook import DictionaryEbook 
from penelope.utilities import create_temp_directory
from penelope.utilities import delete_directory
from penelope.utilities import print_debug
from penelope.utilities import print_error
from penelope.utilities import print_info

__author__ = "Alberto Pettarin"
__copyright__ = "Copyright 2012-2015, Alberto Pettarin (www.albertopettarin.it)"
__license__ = "MIT"
__version__ = "3.1.0"
__email__ = "alberto@albertopettarin.it"
__status__ = "Production"

def read(dictionary, args, input_file_paths):
    print_error("Read function not implemented for EPUB dictionaries")
    return None

def write(dictionary, args, output_file_path):
    # result to be returned
    result = None

    # get absolute path
    output_file_path_absolute = os.path.abspath(output_file_path)

    # sort by headword, optionally ignoring case
    dictionary.sort(by_headword=True, ignore_case=args.sort_ignore_case)

    # create groups
    special_group, group_keys, group_dict = dictionary.group(
        prefix_function_path=args.group_by_prefix_function,
        prefix_length=int(args.group_by_prefix_length),
        merge_min_size=int(args.group_by_prefix_merge_min_size),
        merge_across_first=args.group_by_prefix_merge_across_first
    )
    all_group_keys = group_keys
    if special_group is not None:
        all_group_keys += [u"SPECIAL"]

    # create epub object
    epub = DictionaryEbook(ebook_format=DictionaryEbook.EPUB2, args=args)
   
    # add groups
    for key in all_group_keys:
        if key == u"SPECIAL":
            group_entries = special_group
        else:
            group_entries = group_dict[key]
        epub.add_group(key, group_entries)

    # create output file
    if args.epub_no_compress:
        print_debug("Not compressing the EPUB container")
        epub.write(output_file_path_absolute, compress=False)
    else:
        print_debug("Writing to file '%s'..." % (output_file_path_absolute), args.debug)
        epub.write(output_file_path_absolute, compress=True)
        result = [output_file_path]
        print_debug("Writing to file '%s'... done" % (output_file_path_absolute), args.debug)

    # delete tmp directory
    tmp_path = epub.get_tmp_path()
    if args.epub_no_compress:
        print_info("The uncompressed EPUB is inside dir '%s'" % (tmp_path))
        result = [tmp_path]
    elif args.keep:
        print_info("Not deleting temp dir '%s'" % (tmp_path))
        if result is None:
            result = [tmp_path]
    else:
        epub.delete()
        print_debug("Deleted temp dir '%s'" % (tmp_path), args.debug)

    return result






#!/usr/bin/env python
# coding=utf-8

"""
This is a (trivial) example of an input parser,
just to show the parse() interface.

This parser simply returns the (raw) input dictionary, unchanged.
"""

__author__ = "Alberto Pettarin"
__copyright__ = "Copyright 2012-2015, Alberto Pettarin (www.albertopettarin.it)"
__license__ = "MIT"
__version__ = "3.1.0"
__email__ = "alberto@albertopettarin.it"
__status__ = "Production"

def parse(dictionary, arguments):
    """
    Given the input dictionary and arguments,
    return a (possibly, modified) dictionary.

    The returned dictionary might be the same (input) instance,
    or a new one.

    :param dictionary: the (raw) input dictionary
    :type  dictionary: Dictionary
    :param arguments: the command line arguments
    :type  arguments: Namespace from argparse
    :rtype: Dictionary
    """
    return dictionary






#!/usr/bin/env python
# coding=utf-8

"""
Read/write Kobo dictionaries.

The read function only acquires the index,
as the definition files of the original Kobo dictionaries
are obfuscated/encrypted.

The write function, however, is able to output
fully functional Kobo dictionaries
(provided that their file names match one of the official ones,
probably because they are hard-coded in the Kobo firmware).

The MARISA executables marisa-build and marisa-reverse-lookup
must be installed as the Python module marisa_trie or
as executables in your $PATH.
"""

from __future__ import absolute_import
from io import open
import imp
import gzip
import os
import subprocess
import zipfile

from penelope.prefix_kobo import get_prefix as get_prefix_kobo
from penelope.utilities import create_temp_directory
from penelope.utilities import create_temp_file
from penelope.utilities import delete_directory
from penelope.utilities import delete_file
from penelope.utilities import print_debug
from penelope.utilities import print_error
from penelope.utilities import print_info
from penelope.utilities import rename_file

__author__ = "Alberto Pettarin"
__copyright__ = "Copyright 2012-2015, Alberto Pettarin (www.albertopettarin.it)"
__license__ = "MIT"
__version__ = "3.1.0"
__email__ = "alberto@albertopettarin.it"
__status__ = "Production"

WORDS_FILE_NAME = u"words"
MARISA_BUILD = u"marisa-build"
MARISA_REVERSE_LOOKUP = u"marisa-reverse-lookup"

def read(dictionary, args, input_file_paths):
    def read_single_file(dictionary, args, input_file_path):
        # result flag
        result = False

        # create a tmp file
        tmp_handler, tmp_path = create_temp_file()

        # copy the index file from the zip to the tmp file
        input_file_obj = zipfile.ZipFile(input_file_path)
        tmp_file_obj = open(tmp_path, "wb")
        tmp_file_obj.write(input_file_obj.read(WORDS_FILE_NAME))
        tmp_file_obj.close()
        input_file_obj.close()

        # read index with MARISA
        try:
            # call MARISA with marisa_trie module
            import marisa_trie
            trie = marisa_trie.Trie()
            trie.load(tmp_path)
            for pair in trie.items():
                dictionary.add_entry(headword=pair[0], definition=u"")
            result = True
        except ImportError as exc:
            # call MARISA with subprocess
            print_info("  MARISA cannot be imported as Python module. You might want to install it with:")
            print_info("  $ [sudo] pip install marisa_trie")
            marisa_reverse_lookup_path = MARISA_REVERSE_LOOKUP
            if args.marisa_bin_path is None:
                print_info("  Running '%s' from $PATH" % MARISA_REVERSE_LOOKUP)
            else:
                marisa_reverse_lookup_path = os.path.join(args.marisa_bin_path, MARISA_REVERSE_LOOKUP)
                print_info("  Running '%s' from '%s'" % (MARISA_REVERSE_LOOKUP, args.marisa_bin_path))
            # TODO this is ugly, but it works
            query = (u"\n".join([str(x) for x in range(int(args.marisa_index_size))]) + u"\n").encode("utf-8")
            
            try:
                proc = subprocess.Popen(
                    [marisa_reverse_lookup_path, tmp_path],
                    stdout=subprocess.PIPE,
                    stdin=subprocess.PIPE,
                    stderr=subprocess.PIPE
                )
                stdout = proc.communicate(input=query)[0].decode("utf-8")
                for line in stdout.splitlines():
                    array = line.split("\t")
                    if len(array) >= 2:
                        key = array[1]
                        if args.ignore_case:
                            key = key.lower()
                        dictionary.add_entry(headword=key, definition=u"")
                result = True
            except OSError as exc:
                print_error("  Unable to run '%s' as '%s'" % (MARISA_REVERSE_LOOKUP, marisa_reverse_lookup_path))
                print_error("  Please make sure '%s':" % MARISA_REVERSE_LOOKUP)
                print_error("    1. is available on your $PATH or")
                print_error("    2. specify its path with --marisa-bin-path or")
                print_error("    3. install the marisa_trie Python module")
        except:
            print_debug("Reading from file '%s'... failed" % (input_file_path))

        # delete the tmp file
        delete_file(tmp_handler, tmp_path)
        return result

    for input_file_path in input_file_paths:
        print_debug("Reading from file '%s'..." % (input_file_path), args.debug)
        result = read_single_file(dictionary, args, input_file_path)
        if result:
            print_debug("Reading from file '%s'... success" % (input_file_path), args.debug)
        else:
            print_error("Reading from file '%s'... failed" % (input_file_path))
            return None
    return dictionary

def write(dictionary, args, output_file_path):
    # result to be returned
    result = None

    # get absolute path
    output_file_path_absolute = os.path.abspath(output_file_path)

    # create tmp directory
    cwd = os.getcwd()
    tmp_path = create_temp_directory()
    print_debug("Working in temp dir '%s'" % (tmp_path), args.debug)
    os.chdir(tmp_path)

    # sort by headword
    dictionary.sort(by_headword=True)

    # group by prefix
    files_to_compress = []
    prefix_length = int(args.group_by_prefix_length)
    special_group, group_keys, group_dict = dictionary.group(
        prefix_function=get_prefix_kobo,
        prefix_length=prefix_length,
        merge_min_size=int(args.group_by_prefix_merge_min_size),
        merge_across_first=args.group_by_prefix_merge_across_first
    )
    if special_group is not None:
        special_group_key = u"1" * prefix_length
        group_dict[special_group_key] = special_group
        group_keys = [special_group_key] + group_keys

    # write files
    for key in group_keys:
        # write html file
        file_html_path = key + u".html"
        file_html_obj = open(file_html_path, "wb")
        file_html_obj.write(u"<?xml version=\"1.0\" encoding=\"utf-8\"?><html>".encode("utf-8"))
        for entry in group_dict[key]:
            headword = entry.headword
            definition = entry.definition
            file_html_obj.write((u"<w><a name=\"%s\"/><div><b>%s</b><br/>%s</div></w>" % (headword, headword, definition)).encode("utf-8"))
        file_html_obj.write((u"</html>").encode("utf-8"))
        file_html_obj.close()

        # compress in gz format
        file_html_obj = open(file_html_path, "rb")
        file_gz_path = file_html_path + u".gz"
        file_gz_obj = gzip.open(file_gz_path, "wb")
        file_gz_obj.writelines(file_html_obj)
        file_gz_obj.close()
        file_html_obj.close()

        # delete .html file
        delete_file(None, file_html_path)
        # rename .html.gz file into .html
        rename_file(file_gz_path, file_html_path)
        files_to_compress.append(file_html_path)

    # write words
    file_words_path = WORDS_FILE_NAME
    keys = sorted(dictionary.entries_index.keys())
    try:
        import marisa_trie
        trie = marisa_trie.Trie(keys)
        trie.save(file_words_path)
        result = [file_words_path] 
    except ImportError as exc:
        # call MARISA with subprocess
        print_info("  MARISA cannot be imported as Python module. You might want to install it with:")
        print_info("  $ [sudo] pip install marisa_trie")
        marisa_build_path = MARISA_BUILD
        if args.marisa_bin_path is None:
            print_info("  Running '%s' from $PATH" % MARISA_BUILD)
        else:
            marisa_build_path = os.path.join(args.marisa_bin_path, MARISA_BUILD)
            print_info("  Running '%s' from '%s'" % (MARISA_BUILD, args.marisa_bin_path))
        # TODO this is ugly, but it works
        query = (u"\n".join([x for x in keys]) + u"\n").encode("utf-8")

        try:
            proc = subprocess.Popen(
                [marisa_build_path, "-l", "-o", file_words_path],
                stdout=subprocess.PIPE,
                stdin=subprocess.PIPE,
                stderr=subprocess.PIPE
            )
            proc.communicate(input=query)[0].decode("utf-8")
            result = [file_words_path] 
        except OSError as exc:
            print_error("  Unable to run '%s' as '%s'" % (MARISA_BUILD, marisa_build_path))
            print_error("  Please make sure '%s':" % MARISA_BUILD)
            print_error("    1. is available on your $PATH or")
            print_error("    2. specify its path with --marisa-bin-path or")
            print_error("    3. install the marisa_trie Python module")
            result = None 

    if result is not None:
        # add file_words_path to files to compress
        files_to_compress.append(file_words_path)
        # create output zip file
        try:
            print_debug("Writing to file '%s'..." % (output_file_path_absolute), args.debug)
            file_zip_obj = zipfile.ZipFile(output_file_path_absolute, "w", zipfile.ZIP_DEFLATED)
            for file_to_compress in files_to_compress:
                file_to_compress = os.path.basename(file_to_compress)
                file_zip_obj.write(file_to_compress)
            file_zip_obj.close()
            result = [output_file_path]
            print_debug("Writing to file '%s'... success" % (output_file_path_absolute), args.debug)
        except:
            print_error("Writing to file '%s'... failure" % (output_file_path_absolute))

    # delete tmp directory
    os.chdir(cwd)
    if args.keep:
        print_info("Not deleting temp dir '%s'" % (tmp_path))
    else:
        delete_directory(tmp_path)
        print_debug("Deleted temp dir '%s'" % (tmp_path), args.debug)

    return result






#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This is the prefix function for grouping headwords for Kobo format.
"""

__author__ = "Alberto Pettarin"
__copyright__ = "Copyright 2012-2015, Alberto Pettarin (www.albertopettarin.it)"
__license__ = "MIT"
__version__ = "3.1.0"
__email__ = "alberto@albertopettarin.it"
__status__ = "Production"

def get_prefix(headword, length):
    """
    Return the prefix for the given headword,
    of length length.

    Note that the procedure implemented here is the result
    of reverse engineering, since no official specification
    has been published by Kobo so far. YMMV.

    :param headword: the headword string
    :type  headword: unicode
    :param length: prefix length
    :type  length: int
    :rtype: unicode
    """
    def is_allowed(character):
        # all non-ascii (x > 127) are ok
        # all ASCII lowercase letters (97 <= x <= 122) are ok
        # everything else is not ok
        try:
            code = ord(character)
            return (code > 127) or ((code >= 97) and (code <= 122))
        except:
            pass
        return True

    # defaults to u"SPECIAL", it will be mapped to u"11...1" later
    prefix = u"SPECIAL"
    headword = headword.lower()
    if len(headword) > 0:
        while len(headword) < length:
            # for headwords shorter than length, append an 'a' at the end
            # e.g. length=3, "xy" => "xya"
            headword += u"a"
        # TODO maybe the check should be done only for the first character
        is_ok = True
        for character in headword:
            if not is_allowed(character):
                is_ok = False
                break
        if is_ok:
            prefix = headword[0:length]
    return prefix






#!/usr/bin/env python
# coding=utf-8

"""
Read/write CSV dictionaries.
"""

from __future__ import absolute_import
from io import open

from penelope.utilities import print_debug
from penelope.utilities import print_error

__author__ = "Alberto Pettarin"
__copyright__ = "Copyright 2012-2015, Alberto Pettarin (www.albertopettarin.it)"
__license__ = "MIT"
__version__ = "3.1.0"
__email__ = "alberto@albertopettarin.it"
__status__ = "Production"

def read(dictionary, args, input_file_paths):
    for input_file_path in input_file_paths:
        print_debug("Reading from file '%s'..." % (input_file_path), args.debug)
        input_file_object = open(input_file_path, "rb")
        data_bytes = input_file_object.read() # bytes
        data_unicode = data_bytes.decode(args.input_file_encoding) # unicode
        input_file_object.close()
        lines = data_unicode.split(args.csv_ls)
        if args.csv_ignore_first_line:
            lines = lines[1:]
        for line in lines:
            array = line.split(args.csv_fs)
            if len(array) >= 2:
                headword = array[0]
                definition = line[len(headword) + 1:]
                if args.ignore_case:
                    headword = headword.lower()
                dictionary.add_entry(headword=headword, definition=definition)
        print_debug("Reading from file '%s'... success" % (input_file_path), args.debug)
    return dictionary

def write(dictionary, args, output_file_path):
    try:
        print_debug("Writing to file '%s'..." % (output_file_path), args.debug)
        output_file_obj = open(output_file_path, "wb")
        for index in dictionary.entries_index_sorted:
            entry = dictionary.entries[index]
            string = u"%s%s%s%s" % (
                entry.headword,
                args.csv_fs,
                entry.definition,
                args.csv_ls
            )
            output_file_obj.write(string.encode("utf-8"))
        output_file_obj.close()
        print_debug("Writing to file '%s'... success" % (output_file_path), args.debug)
        return [output_file_path]
    except:
        print_error("Writing to file '%s'... failure" % (output_file_path))
        return None






#!/usr/bin/env python
# coding=utf-8

"""
The Dictionary class is the internal representation of a dictionary.

It is a (multi)dictionary, with a reverse lookup index.

It supports merging definitions for the same headword, and sorting
by headword and/or definition, possibly reversing or ignoring the case.

It also has support for associating synonyms (i.e., headword strings)
to each entry, and to "flatten them" (i.e., creating new headwords,
one for each synonym, with the definition of the original entry).
"""

from __future__ import absolute_import
from io import open
import imp
import os

from penelope.prefix_default import get_prefix as get_prefix_default
from penelope.utilities import get_uuid
from penelope.utilities import print_error

__author__ = "Alberto Pettarin"
__copyright__ = "Copyright 2012-2015, Alberto Pettarin (www.albertopettarin.it)"
__license__ = "MIT"
__version__ = "3.1.0"
__email__ = "alberto@albertopettarin.it"
__status__ = "Production"

def read_dictionary(args):
    """
    Read the input dictionary from the input files specified
    in the given arguments.

    Return a Dictionary, or None if failed.
    """
    metadata = DictionaryMetadata(
        identifier_string=args.identifier,
        author_string=args.author,
        email_string=args.email,
        website_string=args.website,
        title_string=args.title,
        copyright_string=args.copyright,
        license_string=args.license,
        year_string=args.year,
        language_from=args.language_from,
        language_to=args.language_to,
        description_string=args.description
    )
    dictionary = Dictionary(metadata=metadata)
    input_format = args.input_format
    if input_format == "bookeen":
        # NOTE
        # for bookeen format we cannot prepare the file paths
        # as for the other formats, since we might have a compressed .install file
        # or an uncompressed pair (.dict.idx and .dict)
        # this is dealt with directly in format_bookeen.py
        import penelope.format_bookeen
        return penelope.format_bookeen.read(dictionary, args, args.input_file)
    elif input_format == "csv":
        input_file_paths = prepare_file_paths(args.input_file, ".csv")
        if input_file_paths is None:
            return None
        import penelope.format_csv
        return penelope.format_csv.read(dictionary, args, input_file_paths)
    elif input_format == "kobo":
        input_file_paths = prepare_file_paths(args.input_file, ".zip")
        if input_file_paths is None:
            return None
        import penelope.format_kobo
        return penelope.format_kobo.read(dictionary, args, input_file_paths)
    elif input_format == "stardict":
        input_file_paths = prepare_file_paths(args.input_file, ".zip")
        if input_file_paths is None:
            return None
        import penelope.format_stardict
        return penelope.format_stardict.read(dictionary, args, input_file_paths)
    elif input_format == "xml":
        input_file_paths = prepare_file_paths(args.input_file, ".xml")
        if input_file_paths is None:
            return None
        import penelope.format_xml
        return penelope.format_xml.read(dictionary, args, input_file_paths)
    return dictionary

def write_dictionary(dictionary, args):
    """
    Write the dictionary to file.

    Return a list of paths, or None if failed.
    """
    output_format = args.output_format
    if output_format == "bookeen":
        # NOTE
        # for bookeen format we cannot add the file extension
        # as for the other formats, since we might have to generate
        # a compressed .install file or an uncompressed pair (.dict.idx and .dict)
        # this is dealt with directly in format_bookeen.py
        import penelope.format_bookeen
        return penelope.format_bookeen.write(dictionary, args, args.output_file)
    elif output_format == "csv":
        output_file_path = add_extension(args.output_file, ".csv")
        import penelope.format_csv
        return penelope.format_csv.write(dictionary, args, output_file_path)
    elif output_format == "epub":
        output_file_path = add_extension(args.output_file, ".epub")
        import penelope.format_epub
        return penelope.format_epub.write(dictionary, args, output_file_path)
    elif output_format == "kobo":
        output_file_path = get_kobo_file_path(args)
        import penelope.format_kobo
        return penelope.format_kobo.write(dictionary, args, output_file_path)
    elif output_format == "mobi":
        output_file_path = add_extension(args.output_file, ".mobi")
        import penelope.format_mobi
        return penelope.format_mobi.write(dictionary, args, output_file_path)
    elif output_format == "stardict":
        output_file_path = add_extension(args.output_file, ".zip")
        import penelope.format_stardict
        return penelope.format_stardict.write(dictionary, args, output_file_path)
    elif output_format == "xml":
        output_file_path = add_extension(args.output_file, ".xml")
        import penelope.format_xml
        return penelope.format_xml.write(dictionary, args, output_file_path)
    return False

def get_kobo_file_path(args):
    if (args.language_from == args.language_to) and (args.language_from == "en"):
        output_file_name = "dicthtml.zip"
    else:
        output_file_name = "dicthtml-%s" % (args.language_from)
        if args.language_from != args.language_to:
            output_file_name += "-%s" % (args.language_to)
        output_file_name += ".zip"
    output_file_path = os.path.split(args.output_file)[0]
    output_file_path = os.path.join(output_file_path, output_file_name)
    return output_file_path

def add_extension(file_path, extension):
    if not file_path.endswith(extension):
        file_path += extension
    return file_path

def prepare_file_paths(string, extension):
    file_paths = []
    for prefix in string.split(","):
        file_path = add_extension(prefix, extension)
        if not os.path.exists(file_path):
            print_error("File '%s' does not exist" % file_path)
            return None
        file_paths.append(file_path)
    return file_paths



class DictionaryEntry(object):
    def __init__(
            self,
            headword,
            definition
    ):
        self.headword = headword
        self.definition = definition
        self.clear_synonyms()

    def clear_synonyms(self):
        self.synonyms = []

    def add_synonym(self, synonym):
        self.synonyms.append(synonym)

    def get_synonyms(self):
        return self.synonyms

    def __len__(self):
        if self.headword is None:
            return 0
        return len(self.headword)

    def __str__(self):
        return u"""DictionaryEntry
    Headword: '%s'
    Definition: '%s'""" % (self.headword, self.definition)

    def prefix(self, prefix_length):
        if len(self) < prefix_length:
            return self.headword
        return self.headword[0:prefix_length]

class DictionaryMetadata(object):
    def __init__(
            self,
            identifier_string=None,
            author_string=None,
            email_string=None,
            website_string=None,
            title_string=None,
            copyright_string=None,
            license_string=None,
            year_string=None,
            language_from=None,
            language_to=None,
            description_string=None
    ):
        self.identifier_string = identifier_string
        self.author_string = author_string
        self.email_string = email_string
        self.website_string = website_string
        self.title_string = title_string
        self.copyright_string = copyright_string
        self.license_string = license_string
        self.year_string = year_string
        self.language_from = language_from
        self.language_to = language_to
        self.description_string = description_string
        if self.identifier_string is None:
            self.identifier_string = get_uuid()

    def __str__(self):
        return u"""DictionaryMetadata
    Identifier:    '%s'
    Language from: '%s'
    Language to:   '%s'
    Title:         '%s'
    Author:        '%s'
    Copyright:     '%s'
    License:       '%s'
    Year:          '%s'
    Email:         '%s'
    Website:       '%s'
    Description:   '%s'""" % (
        self.identifier_string,
        self.language_from,
        self.language_to,
        self.title_string,
        self.author_string,
        self.copyright_string,
        self.license_string,
        self.year_string,
        self.email_string,
        self.website_string,
        self.description_string
    )

    @property
    def is_monolingual(self):
        if self.language_from is None:
            return False
        return self.language_from == self.language_to

    @property
    def is_bilingual(self):
        if self.language_from is None:
            return False
        return self.language_from != self.language_to

class Dictionary(object):
    def __init__(
            self,
            metadata=None
    ):
        # metadata object
        self.metadata = metadata
        if self.metadata is None:
            self.metadata = DictionaryMetadata()
        # actual dictionary entries
        self.entries = []
        # maps headword -> list of indices of self.entries with that headword
        self.entries_index = {}
        # list of indices (unsorted or sorted by headword and/or definition)
        self.entries_index_sorted = []
        self.has_synonyms = False

    def __str__(self):
        return """Dictionary
Number of entries:          %d
Number of unique headwords: %d
Has synonyms:               %s
%s""" % (len(self), self.unique_headwords, str(self.has_synonyms), str(self.metadata))

    def __len__(self):
        return len(self.entries)

    def clear(self):
        self.entries = []
        self.entries_index = {}
        self.entries_index_sorted = []
        self.has_synonyms = False

    @property
    def unique_headwords(self):
        return len(self.entries_index)

    @property
    def has_unique_headwords_only(self):
        return len(self) == self.unique_headwords

    def add_synonym(self, synonym, headword_index):
        if headword_index < len(self):
            entry = self.entries[headword_index]
            entry.add_synonym(synonym)
            self.has_synonyms = True

    def add_entry(self, entry=None, headword=None, definition=None):
        if entry is None:
            entry = DictionaryEntry(headword, definition)
        self.entries.append(entry)
        if not entry.headword in self.entries_index:
            self.entries_index[entry.headword] = []
        index = len(self.entries) - 1
        self.entries_index[entry.headword].append(index)
        self.entries_index_sorted.append(index)

    def has_headword(self, headword):
        return headword in self.entries_index

    def get_definitions(self, headword):
        if not self.has_headword(headword):
            return []
        definitions = []
        for index in self.entries_index[headword]:
            definitions.append(self.entries[index].definition)
        return definitions

    def get_synonyms(self):
        syn_with_index = []
        if self.has_synonyms:
            for index in self.entries_index_sorted:
                entry = self.entries[index]
                for synonym in entry.get_synonyms():
                    syn_with_index.append([synonym, index])
        return syn_with_index

    def sort(self, by_headword=True, by_definition=False, reverse=False, ignore_case=False):
        if (not by_headword) and (not by_definition):
            self.entries_index_sorted = range(len(self.entries))
            return
        tmp = []
        i = 0
        for entry in self.entries:
            first = entry.headword if by_headword else u""
            if ignore_case:
                first = first.lower()
            second = entry.definition if by_definition else u""
            if ignore_case:
                second = second.lower()
            tmp.append([
                first,
                second,
                i
            ])
            i += 1
        tmp = sorted(tmp, reverse=reverse)
        self.entries_index_sorted = []
        for t in tmp:
            self.entries_index_sorted.append(t[2])

    def flatten_synonyms(self):
        """
        Add a new entry for each synonym,
        using the definition of the original headword.
        At the end, reset the current sort order.
        (You will need to sort it later, if interested in having
        the dictionary sorted by headword and/or definition.)
        """
        if not self.has_synonyms:
            # nothing to do
            return

        for entry in self.entries:
            for synonym in entry.get_synonyms():
                self.add_entry(headword=synonym, definition=entry.definition)

        self.sort(False, False, False, False)

    def merge_definitions(self, merge_function=None, merge_separator=None):
        """
        Merge definitions of entries with the same headword,
        using merge_function or merge_separator to create the merged definition.

        Note that this function destroys the current dictionary,
        and re-populate its entries (with the new "merged" entries).

        As a consequence, it resets the current sort order.
        (You will need to sort it later, if interested in having
        the dictionary sorted by headword and/or definition.)
        """

        def default_merge_function(headword, definitions):
            """
            Merge definitions for the same headword in a custom way:
            1 def   => definition
            2+ defs => definition<SEP>definition<SEP>...
            """
            return merge_separator.join(definitions)

        if (self.has_unique_headwords_only) or ((merge_function is None) and (merge_separator is None)):
            # nothing to do
            return

        if merge_function is None:
            # use the default merge function, joining using the merge_separator string
            merge_function = default_merge_function

        # copy previous data
        original_entries = self.entries
        original_entries_index = self.entries_index
        # delete all
        self.clear()
        # for all (unique) headwords
        for headword in original_entries_index:
            original_entries_index_headword = original_entries_index[headword]
            definitions_to_be_merged = [original_entries[i].definition for i in original_entries_index_headword]
            merged_definition = merge_function(headword, definitions_to_be_merged)
            self.add_entry(headword=headword, definition=merged_definition)
            new_headword_index = len(self) - 1
            # add synonyms to this new entry, by adding all synonyms of the original entries
            # merged into this new entry (with index new_headword_index in self.entries)
            for i in original_entries_index_headword:
                original_entry_synonyms = original_entries[i].get_synonyms()
                for synonym in original_entry_synonyms:
                    self.add_synonym(synonym=synonym, headword_index=new_headword_index)

        # not needed, since we called self.clear()
        #self.sort(False, False, False, False)

    def group(
            self,
            prefix_function=None,
            prefix_function_path=None,
            prefix_length=2,
            merge_min_size=0,
            merge_across_first=False
    ):
        """
        Group headwords by prefix, returning a dictionary containing
        the prefixes as keys (possibly, with a "SPECIAL" key) and
        the dictionary entries as elements of the list associated with a key.

        :param prefix_function_path: the path to a source file containing
                                a get_prefix function, mapping a headword
                                and the prefix_length to the prefix;
                                if None, a default function will be used
        :type  prefix_function_path: path
        :param prefix_function: the function, mapping a headword
                                and the prefix_length to the prefix;
                                if None, a default function will be used
        :type  prefix_function: function
        :param prefix_length: the lenght of the prefixes
        :type  prefix_length: int
        :param merge_min_size: merge headword groups until the given minimum
                               number of headwords is reached; if 0, does not merge
        :type  merge_min_size: int
        :param merge_across_first: if True, merge groups even when
                             the first character changes
        :type  merge_across_first: False
        :rtype: (list, list, dict)
        """
        def return_triple(groups):
            """
            Return a (list_special, list, dict),
            where the list contains the sorted keys of dict,
            and list_special contains the list of SPECIAL entries.
            """
            spec = None
            if u"SPECIAL" in groups:
                spec = groups[u"SPECIAL"]
                del groups[u"SPECIAL"]
            keys = sorted(groups.keys())
            return (spec, keys, groups)

        # load the prefix function
        get_prefix = get_prefix_default
        if prefix_function is not None:
            get_prefix = prefix_function
        elif prefix_function_path is not None:
            try:
                get_prefix = imp.load_source("", prefix_function).get_prefix
            except:
                pass

        # create groups
        raw_groups = {}
        for index in self.entries_index_sorted:
            entry = self.entries[index]
            prefix = get_prefix(entry.headword, prefix_length)
            if not prefix in raw_groups:
                raw_groups[prefix] = []
            raw_groups[prefix].append(self.entries[index])

        # if no merge is requested, return
        if merge_min_size == 0:
            return return_triple(raw_groups)

        # merge small groups
        merged_groups = {}
        if u"SPECIAL" in raw_groups:
            # special is never merged
            merged_groups[u"SPECIAL"] = raw_groups[u"SPECIAL"]
            del raw_groups[u"SPECIAL"]
        keys = sorted(raw_groups.keys())
        accumulator_key = keys[0]
        accumulator = raw_groups[accumulator_key]
        for key in keys[1:]:
            if (
                    (len(accumulator) >= merge_min_size) or
                    ((not merge_across_first) and (key[0] != accumulator_key[0]))
                ):
                merged_groups[accumulator_key] = accumulator
                accumulator_key = key
                accumulator = raw_groups[accumulator_key]
            else:
                accumulator += raw_groups[key]
        merged_groups[accumulator_key] = accumulator
        return return_triple(merged_groups)






#!/usr/bin/env python
# coding=utf-8

"""
Set penelope package up
"""

from setuptools import setup, Extension

__author__ = "Alberto Pettarin"
__copyright__ = "Copyright 2012-2015, Alberto Pettarin (www.albertopettarin.it)"
__license__ = "MIT"
__version__ = "3.1.0"
__email__ = "alberto@albertopettarin.it"
__status__ = "Production"

setup(
    name="penelope",
    packages=["penelope"],
    package_data={"penelope": ["res/*"]},
    version="3.1.0.1",
    description="Penelope is a multi-tool for creating, editing and converting dictionaries, especially for eReader devices",
    author="Alberto Pettarin",
    author_email="alberto@albertopettarin.it",
    url="https://github.com/pettarin/penelope",
    license="MIT License",
    long_description=open("README.rst", "r").read(),
    install_requires=["lxml>=3.0", "marisa-trie>=0.7.2"],
    scripts=["bin/penelope"],
    keywords=[
        "Dictionary",
        "Dictionaries",
        "Index",
        "Merge",
        "Flatten",
        "eReader",
        "eReaders",
        "Bookeen",
        "CSV",
        "EPUB",
        "MOBI",
        "Kindle",
        "Kobo",
        "StarDict",
        "XML",
        "MARISA",
        "kindlegen",
        "dictzip",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Topic :: Desktop Environment",
        "Topic :: Documentation",
        "Topic :: Office/Business",
        "Topic :: Software Development :: Internationalization",
        "Topic :: Software Development :: Localization",
        "Topic :: Text Editors",
        "Topic :: Text Editors :: Text Processing",
        "Topic :: Text Processing",
        "Topic :: Text Processing :: General",
        "Topic :: Text Processing :: Indexing",
        "Topic :: Text Processing :: Linguistic",
        "Topic :: Text Processing :: Markup",
        "Topic :: Text Processing :: Markup :: HTML",
        "Topic :: Text Processing :: Markup :: XML",
        "Topic :: Utilities"
    ],
)



#!/usr/bin/env python
# coding=utf-8

"""
Read/write Bookeen dictionaries.
"""

from __future__ import absolute_import
from io import open
import imp
import os
import sqlite3
import zipfile

from penelope.collation_default import collate_function as collate_function_default
from penelope.utilities import print_debug
from penelope.utilities import print_error
from penelope.utilities import print_info
from penelope.utilities import create_temp_directory
from penelope.utilities import copy_file
from penelope.utilities import delete_directory

__author__ = "Alberto Pettarin"
__copyright__ = "Copyright 2012-2015, Alberto Pettarin (www.albertopettarin.it)"
__license__ = "MIT"
__version__ = "3.1.0"
__email__ = "alberto@albertopettarin.it"
__status__ = "Production"

CHUNK_FILE_PREFIX = "c_"
CHUNK_SIZE = 262144 # 2^18
EMPTY_FILE_PATH = os.path.join(os.path.split(os.path.abspath(__file__))[0], "res/empty.idx")
HEADER = "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\"  \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\" [<!ENTITY ns \"&#8226;\">]><html xml:lang=\"%s\" xmlns=\"http://www.w3.org/1999/xhtml\"><head><title></title></head><body>"

def read(dictionary, args, input_file_string):
    def read_single_dict(dictionary, args, single_dict):
        # create tmp directory
        tmp_path = create_temp_directory()
        print_debug("Working in temp dir '%s'" % (tmp_path), args.debug)

        if len(single_dict) == 1:
            print_debug("Unzipping .install file...", args.debug)
            zip_file_path = single_dict[0]
            idx_file_path = os.path.join(tmp_path, "d.dict.idx")
            dict_file_path = os.path.join(tmp_path, "d.dict")
            zip_file_obj = zipfile.ZipFile(zip_file_path, "r")
            for entry in zip_file_obj.namelist():
                if entry.endswith(".dict.idx"):
                    zip_entry = zip_file_obj.open(entry)
                    idx_file_obj = open(idx_file_path, "wb")
                    idx_file_obj.write(zip_entry.read())
                    idx_file_obj.close()
                    zip_entry.close()
                elif entry.endswith(".dict"):
                    zip_entry = zip_file_obj.open(entry)
                    dict_file_obj = open(dict_file_path, "wb")
                    dict_file_obj.write(zip_entry.read())
                    dict_file_obj.close()
                    zip_entry.close()
            zip_file_obj.close()
            print_debug("Unzipping .install file... done", args.debug)
        else:
            print_debug("Files .dict.idx and .dict already uncompressed...", args.debug)
            idx_file_path = single_dict[0]
            dict_file_path = single_dict[1]
            for file_path in [idx_file_path, dict_file_path]:
                if not os.path.exists(file_path):
                    print_error("File '%s' does not exist" % file_path)
                    return False
            print_debug("Files .dict.idx and .dict already uncompressed... done", args.debug)

        # unzip .dict file into tmp_path
        print_debug("Unzipping .dict file...", args.debug)
        zip_file_obj = zipfile.ZipFile(dict_file_path, "r")
        for entry in zip_file_obj.namelist():
            if not entry.endswith("/"):
                zip_entry = zip_file_obj.open(entry)
                entry_file_path = os.path.join(tmp_path, os.path.basename(entry))
                entry_file_obj = open(entry_file_path, "wb")
                entry_file_obj.write(zip_entry.read())
                entry_file_obj.close()
                zip_entry.close()
        zip_file_obj.close()
        print_debug("Unzipping .dict file... done", args.debug)

        # read .dict.idx
        print_debug("Reading .dict.idx file...", args.debug)
        sql_connection = sqlite3.connect(idx_file_path)
        sql_cursor = sql_connection.cursor()
        sql_cursor.execute("select * from T_DictIndex")
        index_data = sql_cursor.fetchall()
        chunk_index_to_entries = {}
        max_chunk_index = 1
        for index_entry in index_data:
            headword = index_entry[1]
            if args.ignore_case:
                headword = headword.lower()
            offset = index_entry[2]
            size = index_entry[3]
            chunk_index = index_entry[4]
            if not chunk_index in chunk_index_to_entries:
                chunk_index_to_entries[chunk_index] = []
            if chunk_index > max_chunk_index:
                max_chunk_index = chunk_index
            chunk_index_to_entries[chunk_index].append([headword, offset, size])
        sql_cursor.close()
        sql_connection.close()
        print_debug("Reading .dict.idx file... done", args.debug)

        # read c_* files
        print_debug("Reading c_* files...", args.debug)
        for chunk_index in range(1, max_chunk_index + 1):
            print_debug("  Reading c_%d file..." % (chunk_index), args.debug)
            chunk_file_path = os.path.join(tmp_path, "%s%d" % (CHUNK_FILE_PREFIX, chunk_index))
            chunk_file_obj = open(chunk_file_path, "rb")
            for entry in chunk_index_to_entries[chunk_index]:
                headword = entry[0]
                offset = entry[1]
                size = entry[2]
                chunk_file_obj.seek(offset)
                definition_bytes = chunk_file_obj.read(size)
                definition_unicode = definition_bytes.decode(args.input_file_encoding)
                dictionary.add_entry(headword=headword, definition=definition_unicode)
            chunk_file_obj.close()
            print_debug("  Reading c_%d file... done" % (chunk_index), args.debug)
        print_debug("Reading c_* files... done", args.debug)

        # delete tmp directory
        if args.keep:
            print_info("Not deleting temp dir '%s'" % (tmp_path))
        else:
            delete_directory(tmp_path)
            print_debug("Deleted temp dir '%s'" % (tmp_path), args.debug)
        return True 

    single_dicts = []
    for prefix in input_file_string.split(","):
        if prefix.endswith(".install"):
            single_dicts.append([prefix])
        elif prefix.endswith(".dict"):
            tentative_dict_path = prefix
            tentative_idx_path = tentative_dict_path + u".idx"
            if (os.path.exists(tentative_idx_path)) and (os.path.exists(tentative_dict_path)):
                single_dicts.append([tentative_idx_path, tentative_dict_path])
        else:
            tentative_dict_path = prefix + u".dict"
            tentative_idx_path = tentative_dict_path + u".idx"
            if (os.path.exists(tentative_idx_path)) and (os.path.exists(tentative_dict_path)):
                single_dicts.append([tentative_idx_path, tentative_dict_path])
    
    if len(single_dicts) == 0:
        print_error("Cannot find .install or .dict.idx/.dict files")
        return None

    for single_dict in single_dicts:
        print_debug("Reading from file '%s'..." % (single_dict), args.debug)
        result = read_single_dict(dictionary, args, single_dict)
        if result:
            print_debug("Reading from file '%s'... success" % (single_dict), args.debug)
        else:
            print_error("Reading from file '%s'... failed" % (single_dict))
            return None
    return dictionary

def write(dictionary, args, output_file_path):
    # result to be returned
    result = None

    # get absolute path
    output_file_path_absolute = os.path.abspath(output_file_path)

    # get absolute path for collation function file 
    bookeen_collation_function_path = None
    if args.bookeen_collation_function is not None:
        bookeen_collation_function_path = os.path.abspath(args.bookeen_collation_function)

    # create tmp directory
    cwd = os.getcwd()
    tmp_path = create_temp_directory()
    print_debug("Working in temp dir '%s'" % (tmp_path), args.debug)
    os.chdir(tmp_path)

    # get the basename
    base = os.path.basename(output_file_path)
    if base.endswith(".zip"):
        base = base[:-4]

    # copy empty.idx into tmp_path
    idx_file_path = base + u".dict.idx"
    dict_file_path = base + u".dict"
    copy_file(EMPTY_FILE_PATH, idx_file_path)

    # open index
    sql_connection = sqlite3.connect(idx_file_path)

    # install collation in the index
    collation_function = collate_function_default
    if bookeen_collation_function_path is not None:
        try:
            collation_function = imp.load_source("", bookeen_collation_function_path).collate_function
            print_debug("Using collation function from '%s'" % (bookeen_collation_function_path), args.debug)
        except:
            print_error("Unable to load collation function from '%s'. Using the default collation function instead." % (bookeen_collation_function_path))
    sql_connection.create_collation("IcuNoCase", collation_function)
    sql_connection.text_factory = str

    # get a cursor and delete any data from the index file
    sql_cursor = sql_connection.cursor()
    sql_cursor.execute("delete from T_DictIndex")

    # write c_* files
    # each c_* file has MAX_CHUNK_SIZE < size <= (MAX_CHUNK_SIZE * 2) bytes (tentatively)
    print_debug("Writing c_* files...", args.debug)
    files_to_compress = []
    current_offset = 0
    chunk_index = 1
    chunk_file_path = "%s%d" % (CHUNK_FILE_PREFIX, chunk_index)
    files_to_compress.append(chunk_file_path)
    chunk_file_obj = open(chunk_file_path, "wb")
    for entry_index in dictionary.entries_index_sorted:
        entry = dictionary.entries[entry_index]
        definition_bytes = entry.definition.encode("utf-8")
        definition_size = len(definition_bytes)
        chunk_file_obj.write(definition_bytes)
        # insert headword into index file
        sql_tuple = (0, entry.headword, current_offset, definition_size, chunk_index)
        sql_cursor.execute("insert into T_DictIndex values (?,?,?,?,?)", sql_tuple)
        # insert synonyms into index file
        if not args.ignore_synonyms:
            for synonym in entry.get_synonyms():
                sql_tuple = (0, synonym[0], current_offset, definition_size, chunk_index)
                sql_cursor.execute("insert into T_DictIndex values (?,?,?,?,?)", sql_tuple)
        # update offset
        current_offset += definition_size
        # if we reached CHUNK_SIZE, open the next c_* file
        if current_offset > CHUNK_SIZE:
            chunk_file_obj.close()
            chunk_index += 1
            chunk_file_path = "%s%d" % (CHUNK_FILE_PREFIX, chunk_index)
            files_to_compress.append(chunk_file_path)
            chunk_file_obj = open(chunk_file_path, "wb")
            current_offset = 0
    chunk_file_obj.close()
    print_debug("Writing c_* files... done", args.debug)

    # compress
    print_debug("Compressing c_* files...", args.debug)
    file_zip_obj = zipfile.ZipFile(dict_file_path, "w", zipfile.ZIP_DEFLATED)
    for file_to_compress in files_to_compress:
        file_to_compress = os.path.basename(file_to_compress)
        file_zip_obj.write(file_to_compress)
    file_zip_obj.close()
    print_debug("Compressing c_* files... done", args.debug)

    # update index metadata
    print_debug("Updating index metadata...", args.debug)
    header = HEADER % (args.language_from)
    sql_cursor.execute("update T_DictInfo set F_xhtmlHeader=?", (header,))
    sql_cursor.execute("update T_DictInfo set F_LangFrom=?", (args.language_from,))
    sql_cursor.execute("update T_DictInfo set F_LangTo=?", (args.language_to,))
    sql_cursor.execute("update T_DictInfo set F_Licence=?", (args.license,))
    sql_cursor.execute("update T_DictInfo set F_Copyright=?", (args.copyright,))
    sql_cursor.execute("update T_DictInfo set F_Title=?", (args.title,))
    sql_cursor.execute("update T_DictInfo set F_Description=?", (args.description,))
    sql_cursor.execute("update T_DictInfo set F_Year=?", (args.year,))
    # the meaning of the following is unknown 
    sql_cursor.execute("update T_DictInfo set F_Alphabet=?", ("Z",))
    sql_cursor.execute("update T_DictInfo set F_CollationLevel=?", ("1",))
    sql_cursor.execute("update T_DictVersion set F_DictType=?", ("stardict",))
    sql_cursor.execute("update T_DictVersion set F_Version=?", ("11",))
    print_debug("Updating index metadata... done", args.debug)

    # compact and close
    sql_cursor.execute("vacuum")
    sql_cursor.close()
    sql_connection.close()

    # create .install file or copy .dict.idx and .dict into requested output directory
    parent_output_directory = os.path.split(output_file_path_absolute)[0]
    if args.bookeen_install_file:
        print_debug("Creating .install file...", args.debug)
        file_zip_path = os.path.join(parent_output_directory, base + u".install")
        file_zip_obj = zipfile.ZipFile(file_zip_path, "w", zipfile.ZIP_DEFLATED)
        for file_to_compress in [dict_file_path, idx_file_path]:
            file_to_compress = os.path.basename(file_to_compress)
            file_zip_obj.write(file_to_compress)
        file_zip_obj.close()
        result = [file_zip_path]
        print_debug("Creating .install file... done", args.debug)
    else:
        print_debug("Copying .dict.idx and .dict files...", args.debug)
        dict_file_path_final = os.path.join(parent_output_directory, os.path.basename(dict_file_path))
        idx_file_path_final = os.path.join(parent_output_directory, os.path.basename(idx_file_path))
        copy_file(dict_file_path, dict_file_path_final)
        copy_file(idx_file_path, idx_file_path_final)
        result = [idx_file_path_final, dict_file_path_final]
        print_debug("Copying .dict.idx and .dict files... done", args.debug)

    # delete tmp directory
    os.chdir(cwd)
    if args.keep:
        print_info("Not deleting temp dir '%s'" % (tmp_path))
    else:
        delete_directory(tmp_path)
        print_debug("Deleted temp dir '%s'" % (tmp_path), args.debug)

    return result






#!/usr/bin/env python
# coding=utf-8

"""
Write MOBI (Kindle) dictionaries.

It creates a .opf and .html file, and invokes kindlegen to compile it.
"""

from __future__ import absolute_import
from io import open
import os
import subprocess

from penelope.dictionary_ebook import DictionaryEbook
from penelope.utilities import print_debug
from penelope.utilities import print_error
from penelope.utilities import print_info
from penelope.utilities import create_temp_directory
from penelope.utilities import copy_file
from penelope.utilities import delete_directory

__author__ = "Alberto Pettarin"
__copyright__ = "Copyright 2012-2015, Alberto Pettarin (www.albertopettarin.it)"
__license__ = "MIT"
__version__ = "3.1.0"
__email__ = "alberto@albertopettarin.it"
__status__ = "Production"

KINDLEGEN = u"kindlegen"

def read(dictionary, args, input_file_paths):
    print_error("Read function not implemented for MOBI dictionaries")
    return None

def write(dictionary, args, output_file_path):
    # result to be returned
    result = None

    # get absolute path
    output_file_path_absolute = os.path.abspath(output_file_path)

    # sort by headword, optionally ignoring case
    dictionary.sort(by_headword=True, ignore_case=args.sort_ignore_case)

    # create groups
    special_group, group_keys, group_dict = dictionary.group(
        prefix_function_path=args.group_by_prefix_function,
        prefix_length=int(args.group_by_prefix_length),
        merge_min_size=int(args.group_by_prefix_merge_min_size),
        merge_across_first=args.group_by_prefix_merge_across_first
    )
    all_group_keys = group_keys
    if special_group is not None:
        all_group_keys += [u"SPECIAL"]

    # create mobi object
    mobi = DictionaryEbook(ebook_format=DictionaryEbook.MOBI, args=args)

    # add groups
    for key in all_group_keys:
        if key == u"SPECIAL":
            group_entries = special_group
        else:
            group_entries = group_dict[key]
        mobi.add_group(key, group_entries)

    # create output file
    print_debug("Writing to file '%s'..." % (output_file_path_absolute), args.debug)
    mobi.write(output_file_path_absolute, compress=False)
    result = [output_file_path]
    print_debug("Writing to file '%s'... done" % (output_file_path_absolute), args.debug)

    # run kindlegen
    tmp_path = mobi.get_tmp_path()
    if args.mobi_no_kindlegen:
        print_info("Not running kindlegen, the raw files are located in '%s'" % tmp_path)
        result = [tmp_path]
    else:
        try:
            print_debug("Creating .mobi file with kindlegen...", args.debug)
            kindlegen_path = KINDLEGEN
            opf_file_path_absolute = os.path.join(tmp_path, "OEBPS", "content.opf")
            mobi_file_path_relative = u"content.mobi"
            mobi_file_path_absolute = os.path.join(tmp_path, "OEBPS", mobi_file_path_relative)
            if args.kindlegen_path is None:
                print_info("  Running '%s' from $PATH" % KINDLEGEN)
            else:
                kindlegen_path = args.kindlegen_path
                print_info("  Running '%s' from '%s'" % (KINDLEGEN, kindlegen_path))
            proc = subprocess.Popen(
                [kindlegen_path, opf_file_path_absolute, "-o", mobi_file_path_relative],
                stdout=subprocess.PIPE,
                stdin=subprocess.PIPE,
                stderr=subprocess.PIPE
            )
            output = proc.communicate()
            if args.debug:
                output_unicode = (output[0]).decode("utf-8")
                print_debug(output_unicode, args.debug)
            copy_file(mobi_file_path_absolute, output_file_path_absolute)
            result = [output_file_path]
            print_debug("Creating .mobi file with kindlegen... done", args.debug)
        except OSError as exc:
            print_error("  Unable to run '%s' as '%s'" % (KINDLEGEN, kindlegen_path))
            print_error("  Please make sure '%s':" % KINDLEGEN)
            print_error("    1. is available on your $PATH or")
            print_error("    2. specify its path with --kindlegen-path")

    # delete tmp directory
    tmp_path = mobi.get_tmp_path()
    if args.keep:
        print_info("Not deleting temp dir '%s'" % (tmp_path))
    else:
        mobi.delete()
        print_debug("Deleted temp dir '%s'" % (tmp_path), args.debug)

    return result






#!/usr/bin/env python
# coding=utf-8

"""
Read/write XML dictionaries.
"""

from __future__ import absolute_import
from io import open
from lxml import etree

from penelope.utilities import print_debug
from penelope.utilities import print_error

__author__ = "Alberto Pettarin"
__copyright__ = "Copyright 2012-2015, Alberto Pettarin (www.albertopettarin.it)"
__license__ = "MIT"
__version__ = "3.1.0"
__email__ = "alberto@albertopettarin.it"
__status__ = "Production"

def read(dictionary, args, input_file_paths):
    for input_file_path in input_file_paths:
        print_debug("Reading from file '%s'..." % (input_file_path), args.debug)
        input_file_object = open(input_file_path, "rb")
        data_bytes = input_file_object.read() # bytes
        input_file_object.close()
        root = etree.fromstring(data_bytes)
        for entry in root.iter("entry"):
            headword = None
            definition = None
            for child in entry:
                if child.tag == "key":
                    headword = child.text
                if child.tag == "def":
                    definition = child.text
            if (headword is not None) and (definition is not None):    
                dictionary.add_entry(headword=headword, definition=definition)
        print_debug("Reading from file '%s'... success" % (input_file_path), args.debug)
    return dictionary

def write(dictionary, args, output_file_path):
    try:
        print_debug("Creating XML tree...", args.debug)
        dictionary_elem = etree.Element("dictionary")
        for index in dictionary.entries_index_sorted:
            entry = dictionary.entries[index]
            entry_elem = etree.SubElement(dictionary_elem, "entry")
            key_elem = etree.SubElement(entry_elem, "key")
            key_elem.text = entry.headword
            def_elem = etree.SubElement(entry_elem, "def")
            def_elem.text = entry.definition
        tree = etree.ElementTree(dictionary_elem)
        print_debug("Creating XML tree... done", args.debug)
        print_debug("Writing to file '%s'..." % (output_file_path), args.debug)
        tree.write(
            output_file_path,
            pretty_print=True,
            xml_declaration=True
        )
        print_debug("Writing to file '%s'... success" % (output_file_path), args.debug)
        return [output_file_path]
    except:
        print_error("Writing to file '%s'... failure" % (output_file_path))
        return None






#!/usr/bin/env python
# coding=utf-8

"""
Read/write StarDict dictionaries.

To write a StarDict file, the dictzip executable is required.
"""

from __future__ import absolute_import
from io import open
import gzip
import os
import subprocess
import struct
import zipfile

from penelope.utilities import create_temp_directory
from penelope.utilities import delete_directory
from penelope.utilities import print_debug
from penelope.utilities import print_error
from penelope.utilities import print_info

__author__ = "Alberto Pettarin"
__copyright__ = "Copyright 2012-2015, Alberto Pettarin (www.albertopettarin.it)"
__license__ = "MIT"
__version__ = "3.1.0"
__email__ = "alberto@albertopettarin.it"
__status__ = "Production"

DICTZIP = u"dictzip"

SAMETYPESEQUENCE_SUPPORTED_VALUES = [
    u"m", # pure text
    u"l", # pure text (locale encoding instead of UTF-8, use --input-file-encoding to specify it)
    u"g", # Pango markup
    u"t", # English phonetic
    u"x", # XDXF markup
    u"y", # Chinese YinBiao or Japanese KANA
    u"k", # KingSoft PowerWord markup
    u"w", # MediaWiki markup
    u"h"  # HTML markup
]         # all the above are UTF-8 encoded (except "l") and terminated by \0

def read(dictionary, args, input_file_paths):
    def find_files(entries):
        found = {}
        for entry in entries:
            if entry.endswith(".ifo"):
                found["d.ifo"] = entry
                break
        if not "d.ifo" in found:
            print_error("Cannot find .ifo file in the given StarDict file (see StarDict spec)")
            return {}
        # remove .ifo extension
        base = found["d.ifo"][:-4]
        # attempt to find these ones
        tentative_idx = base + ".idx"
        tentative_idx_gz = base + ".idx.gz"
        tentative_dict = base + ".dict"
        tentative_dict_dz = base + ".dict.dz"
        tentative_dz = base + ".dz"
        if tentative_idx in entries:
            found["d.idx"] = tentative_idx
        if tentative_idx_gz in entries:
            found["d.idx.gz"] = tentative_idx_gz
        if not (("d.idx" in found) or ("d.idx.gz" in found)):
            print_error("Cannot find .idx or .idx.gz file in the given StarDict file (see StarDict spec)")
            return {}
        if tentative_dict in entries:
            found["d.dict"] = tentative_dict
        if tentative_dict_dz in entries:
            found["d.dict.dz"] = tentative_dict_dz
        if tentative_dz in entries:
            found["d.dz"] = tentative_dz
        if not (("d.dict" in found) or ("d.dict.dz" in found) or ("d.dz" in found)):
            print_error("Cannot find .dict, .dict.dz, or .dz file in the given StarDict file (see StarDict spec)")
            return {}
        # syn is optional
        tentative_syn = base + ".syn"
        if tentative_syn in entries:
            found["d.syn"] = tentative_syn
        return found

    def uncompress_file(compressed_path, tmp_path, key):
        uncompressed_path = os.path.join(tmp_path, key)
        u_obj = open(uncompressed_path, "wb")
        c_obj = gzip.open(compressed_path, "rb")
        u_obj.write(c_obj.read())
        c_obj.close()
        u_obj.close()
        print_debug("Uncompressed %s" % (uncompressed_path), args.debug)
        return uncompressed_path

    def read_ifo(ifo_path, has_syn, args):
        ifo_dict = {}
        ifo_obj = open(ifo_path, "rb")
        ifo_bytes = ifo_obj.read() # bytes
        ifo_unicode = ifo_bytes.decode("utf-8") # unicode, always utf-8 by spec
        ifo_obj.close()
        for line in ifo_unicode.splitlines():
            array = line.split("=")
            if len(array) >= 2:
                key = array[0]
                val = "=".join(array[1:])
                ifo_dict[key] = val
        
        if not "version" in ifo_dict:
            print_error("No 'version' found in the .ifo file (see StarDict spec)")
            return None
        if ifo_dict["version"] not in ["2.4.2", "3.0.0"]:
            print_error("The .ifo file must have a 'version' value equal to '2.4.2' or '3.0.0' (see StarDict spec)")
            return None

        required_keys = ["bookname", "wordcount", "idxfilesize"]
        if has_syn:
            required_keys.append("synwordcount")
        # TODO not used => disabling this
        #if ifo_dict["version"] == "3.0.0":
        #    required_keys.append("idxoffsetbits")
        for key in required_keys:
            if not key in ifo_dict:
                print_error("No '%s' found in the .ifo file (see StarDict spec)" % key)
                return None

        ifo_dict["wordcount"] = int(ifo_dict["wordcount"])
        ifo_dict["idxfilesize"] = int(ifo_dict["idxfilesize"])
        if has_syn:
            ifo_dict["synwordcount"] = int(ifo_dict["synwordcount"])
        # TODO not used => disabling this
        #if ifo_dict["version"] == "3.0.0":
        #    ifo_dict["idxoffsetbits"] = int(ifo_dict["idxoffsetbits"])

        if args.sd_ignore_sametypesequence:
            print_debug("Ignoring sametypesequence value", args.debug)
        else:
            # TODO limitation: we require sametypesequence to be present
            if not "sametypesequence" in ifo_dict:
                print_error("The .ifo file must have a 'sametypesequence' value (see README).")
                return None
            # TODO limitation: we require sametypesequence to have a value in SAMETYPESEQUENCE_SUPPORTED_VALUES
            if not ifo_dict["sametypesequence"] in SAMETYPESEQUENCE_SUPPORTED_VALUES:
                print_error("The .ifo file must have a 'sametypesequence' value of %s (see README)." % "|".join(SAMETYPESEQUENCE_SUPPORTED_VALUES))
                return None

        return ifo_dict

    def read_single_file(dictionary, args, input_file_path):
        # result flag
        result = False

        # create a tmp directory
        tmp_path = create_temp_directory()
        print_debug("Working in temp dir '%s'" % (tmp_path), args.debug)

        # find .ifo, .idx, .dict[.dz] and .syn files inside the zip
        # and extract them to tmp_path
        input_file_obj = zipfile.ZipFile(input_file_path)
        found_files = find_files(input_file_obj.namelist())
        extracted_files = {}
        if len(found_files) > 0:
            for key in found_files:
                entry = found_files[key]
                ext_file_path = os.path.join(tmp_path, key)
                ext_file_obj = open(ext_file_path, "wb")
                zip_entry = input_file_obj.open(entry)
                ext_file_obj.write(zip_entry.read())
                zip_entry.close()
                ext_file_obj.close()
                print_debug("Extracted %s" % (ext_file_path), args.debug)
                extracted_files[key] = ext_file_path
                # extract from compressed file, but only if ".idx" is not present as well
                if (key == "d.idx.gz") and ("d.idx" not in found_files):
                    extracted_files["d.idx"] = uncompress_file(ext_file_path, tmp_path, "d.idx")
                # extract from compressed file, but only if ".dict" is not present as well
                if ((key == "d.dict.dz") or (key == "d.dz")) and ("d.dict" not in found_files):
                    extracted_files["d.dict"] = uncompress_file(ext_file_path, tmp_path, "d.dict")
        input_file_obj.close()

        # here we have d.ifo, d.idx and d.dict (all uncompressed) and possibly d.syn

        has_syn = "d.syn" in extracted_files
        if (has_syn) and (args.ignore_synonyms):
            has_syn = False
            print_debug("Dictionary has synonyms, but ignoring them (--ignore-synonym)", args.debug)
        ifo_dict = read_ifo(extracted_files["d.ifo"], has_syn, args)
        print_debug("Read .ifo file with values:\n%s" % (str(ifo_dict)), args.debug)

        # read dict file
        dict_file_obj = open(extracted_files["d.dict"], "rb")
        dict_file_bytes = dict_file_obj.read()
        dict_file_obj.close()

        # read idx file
        idx_file_obj = open(extracted_files["d.idx"], "rb")
        byte_read = idx_file_obj.read(1)
        headword = b""
        while byte_read:
            if byte_read == b"\0":
                # end of current word: read offset and size
                offset_bytes = idx_file_obj.read(4)
                offset_int = int((struct.unpack('>i', offset_bytes))[0])
                size_bytes = idx_file_obj.read(4)
                size_int = int((struct.unpack('>i', size_bytes))[0])
                definition = dict_file_bytes[offset_int:offset_int+size_int].decode(args.input_file_encoding)
                headword = headword.decode("utf-8")
                if args.ignore_case:
                    headword = headword.lower()
                dictionary.add_entry(headword=headword, definition=definition)
                headword = b""
            else:
                # read next byte
                headword += byte_read
            byte_read = idx_file_obj.read(1)
        idx_file_obj.close()
        result = True

        # read syn file, if present
        if has_syn:
            print_debug("The input StarDict file contains a .syn file, parsing it...", args.debug)
            result = False
            syn_file_obj = open(extracted_files["d.syn"], "rb")
            byte_read = syn_file_obj.read(1)
            synonym = b""
            while byte_read:
                if byte_read == b"\0":
                    # end of current synonym: read index of original word
                    index_bytes = syn_file_obj.read(4)
                    index_int = int((struct.unpack('>i', index_bytes))[0])
                    synonym = synonym.decode("utf-8")
                    if index_int < len(dictionary):
                        dictionary.add_synonym(synonym=synonym, headword_index=index_int)
                    else:
                        # emit a warning?
                        print_debug("Synonym '%s' points to index %d >= len(dictionary), skipping it" % (index_int, synonym), args.debug)
                    synonym = b""
                else:
                    # read next byte
                    synonym += byte_read
                byte_read = syn_file_obj.read(1)
            syn_file_obj.close()
            result = True
            print_debug("The input StarDict file contains a .syn file, parsing it... done", args.debug)
        else:
            print_debug("The input StarDict file does not contain a .syn file", args.debug)

        # delete tmp directory
        if args.keep:
            print_info("Not deleting temp dir '%s'" % (tmp_path))
        else:
            delete_directory(tmp_path)
            print_debug("Deleted temp dir '%s'" % (tmp_path), args.debug)

        return result

    for input_file_path in input_file_paths:
        print_debug("Reading from file '%s'..." % (input_file_path), args.debug)
        result = read_single_file(dictionary, args, input_file_path)
        if result:
            print_debug("Reading from file '%s'... success" % (input_file_path), args.debug)
        else:
            print_error("Reading from file '%s'... failed" % (input_file_path))
            return None
    return dictionary

def write(dictionary, args, output_file_path):
    # result to be returned
    result = None

    # get absolute path
    output_file_path_absolute = os.path.abspath(output_file_path)

    # create tmp directory
    cwd = os.getcwd()
    tmp_path = create_temp_directory()
    print_debug("Working in temp dir '%s'" % (tmp_path), args.debug)
    os.chdir(tmp_path)

    # get the basename and compute output file paths
    base = os.path.basename(output_file_path)
    if base.endswith(".zip"):
        base = base[:-4]
    ifo_file_path = base + ".ifo"
    idx_file_path = base + ".idx"
    dict_file_path = base + ".dict"
    dict_dz_file_path = base + ".dict.dz"
    syn_file_path = base + ".syn"

    # TODO by spec, the index should be sorted
    # TODO using the comparator stardict_strcmp() defined in the spec
    # TODO (it calls g_ascii_strcasecmp() and/or strcmp() ),
    # TODO or with a user-defined collation function
    #
    # From https://developer.gnome.org/glib/2.28/glib-String-Utility-Functions.html#g-ascii-strcasecmp
    # gint g_ascii_strcasecmp (const gchar *s1, const gchar *s2);
    # Compare two strings, ignoring the case of ASCII characters.
    # Unlike the BSD strcasecmp() function, this only recognizes standard ASCII letters and ignores the locale, treating all non-ASCII bytes as if they are not letters.
    # This function should be used only on strings that are known to be in encodings where the bytes corresponding to ASCII letters always represent themselves. This includes UTF-8 and the ISO-8859-* charsets, but not for instance double-byte encodings like the Windows Codepage 932, where the trailing bytes of double-byte characters include all ASCII letters. If you compare two CP932 strings using this function, you will get false matches. 
    #
    # using Python's builtin lower() and sort() by headword
    # should be equivalent for UTF-8 encoded dictionaries (and it is fast)
    #
    dictionary.sort(by_headword=True, ignore_case=True)

    # write .idx and .dict files
    print_debug("Writing .idx and .dict files...", args.debug)
    idx_file_obj = open(idx_file_path, "wb")
    dict_file_obj = open(dict_file_path, "wb")
    current_offset = 0
    current_idx_size = 0
    for entry_index in dictionary.entries_index_sorted:
        entry = dictionary.entries[entry_index]
        headword_bytes = entry.headword.encode("utf-8")
        definition_bytes = entry.definition.encode("utf-8")
        definition_size = len(definition_bytes)
        # write .idx
        idx_file_obj.write(headword_bytes)
        idx_file_obj.write(b"\0")
        idx_file_obj.write(struct.pack('>i', current_offset))
        idx_file_obj.write(struct.pack('>i', definition_size))
        current_idx_size += (len(headword_bytes) + 1 + 4 + 4)
        # write .dict
        dict_file_obj.write(definition_bytes)
        current_offset += definition_size
    idx_file_obj.close()
    dict_file_obj.close()
    print_debug("Writing .idx and .dict files... done", args.debug)

    # list files to compress
    files_to_compress = []
    files_to_compress.append(ifo_file_path)
    files_to_compress.append(idx_file_path)

    # write .syn file
    dict_syns_len = 0
    if dictionary.has_synonyms:
        if args.ignore_synonyms:
            print_debug("Dictionary has synonyms, but ignoring them", args.debug)
        else:
            print_debug("Dictionary has synonyms, writing .syn file...", args.debug)
            syn_file_obj = open(syn_file_path, "wb")
            dict_syns = dictionary.get_synonyms()
            dict_syns_len = len(dict_syns)
            for pair in dict_syns:
                synonym_bytes = pair[0].encode("utf-8")
                index = pair[1]
                syn_file_obj.write(synonym_bytes)
                syn_file_obj.write(b"\0")
                syn_file_obj.write(struct.pack('>i', index))
            syn_file_obj.close()
            files_to_compress.append(syn_file_path)
            print_debug("Dictionary has synonyms, writing .syn file... done", args.debug)

    # compress .dict file
    if args.sd_no_dictzip:
        print_debug("Not compressing .dict file with dictzip", args.debug)
        files_to_compress.append(dict_file_path)
        result = [dict_file_path] 
    else:
        try:
            print_debug("Compressing .dict file with dictzip...", args.debug)
            dictzip_path = DICTZIP
            if args.dictzip_path is None:
                print_info("  Running '%s' from $PATH" % DICTZIP)
            else:
                dictzip_path = args.dictzip_path
                print_info("  Running '%s' from '%s'" % (DICTZIP, dictzip_path))
            proc = subprocess.Popen(
                [dictzip_path, "-k", dict_file_path],
                stdout=subprocess.PIPE,
                stdin=subprocess.PIPE,
                stderr=subprocess.PIPE
            )
            proc.communicate()
            result = [dict_dz_file_path] 
            files_to_compress.append(dict_dz_file_path)
            print_debug("Compressing .dict file with dictzip... done", args.debug)
        except OSError as exc:
            print_error("  Unable to run '%s' as '%s'" % (DICTZIP, dictzip_path))
            print_error("  Please make sure '%s':" % DICTZIP)
            print_error("    1. is available on your $PATH or")
            print_error("    2. specify its path with --dictzip-path or")
            print_error("    3. specify --no-dictzip to avoid compressing the .dict file")
            result = None 

    if result is not None:
        # create ifo file
        ifo_file_obj = open(ifo_file_path, "wb")
        ifo_file_obj.write((u"StarDict's dict ifo file\n").encode("utf-8"))
        ifo_file_obj.write((u"version=2.4.2\n").encode("utf-8"))
        ifo_file_obj.write((u"wordcount=%d\n" % (len(dictionary))).encode("utf-8"))
        ifo_file_obj.write((u"idxfilesize=%d\n" % (current_idx_size)).encode("utf-8"))
        ifo_file_obj.write((u"bookname=%s\n" % (args.title)).encode("utf-8"))
        ifo_file_obj.write((u"date=%s\n" % (args.year)).encode("utf-8"))
        ifo_file_obj.write((u"sametypesequence=m\n").encode("utf-8"))
        ifo_file_obj.write((u"description=%s\n" % (args.description)).encode("utf-8"))
        ifo_file_obj.write((u"author=%s\n" % (args.author)).encode("utf-8"))
        ifo_file_obj.write((u"email=%s\n" % (args.email)).encode("utf-8"))
        ifo_file_obj.write((u"website=%s\n" % (args.website)).encode("utf-8"))
        if dict_syns_len > 0:
            ifo_file_obj.write((u"synwordcount=%d\n" % (dict_syns_len)).encode("utf-8"))
        ifo_file_obj.close()

        # create output zip file
        try:
            print_debug("Writing to file '%s'..." % (output_file_path_absolute), args.debug)
            file_zip_obj = zipfile.ZipFile(output_file_path_absolute, "w", zipfile.ZIP_DEFLATED)
            for file_to_compress in files_to_compress:
                file_to_compress = os.path.basename(file_to_compress)
                file_zip_obj.write(file_to_compress)
                print_debug("Written %s" % (file_to_compress), args.debug)
            file_zip_obj.close()
            result = [output_file_path]
            print_debug("Writing to file '%s'... success" % (output_file_path_absolute), args.debug)
        except:
            print_error("Writing to file '%s'... failure" % (output_file_path_absolute))

    # delete tmp directory
    os.chdir(cwd)
    if args.keep:
        print_info("Not deleting temp dir '%s'" % (tmp_path))
    else:
        delete_directory(tmp_path)
        print_debug("Deleted temp dir '%s'" % (tmp_path), args.debug)

    return result






#!/usr/bin/env python
# coding=utf-8

"""
DictionaryEbook represents a dictionary ebook
in EPUB 2 and MOBI format.
"""

from __future__ import absolute_import
from __future__ import print_function
from io import open
import os
import zipfile

from penelope.utilities import create_temp_directory
from penelope.utilities import delete_directory

__author__ = "Alberto Pettarin"
__copyright__ = "Copyright 2012-2015, Alberto Pettarin (www.albertopettarin.it)"
__license__ = "MIT"
__version__ = "3.1.0"
__email__ = "alberto@albertopettarin.it"
__status__ = "Production"

class DictionaryEbook():
    """
    A class representing a generic ebook containing a dictionary.

    It can be used to output a MOBI or an EPUB 2 container.

    The ebook must have an OPF, and one or more group XHTML files.

    Optionally, it can have a cover image, an NCX TOC, an index XHTML file.

    The actual file templates are provided by the caller.
    """

    EPUB2 = u"epub2"

    #EPUB3 = u"epub3"

    MOBI = u"mobi"

    GROUP_START_INDEX = 2

    MIMETYPE_CONTENTS = u"application/epub+zip"

    CONTAINER_XML_CONTENTS = u"""<?xml version="1.0" encoding="UTF-8" ?>
<container version="1.0" xmlns="urn:oasis:names:tc:opendocument:xmlns:container">
   <rootfiles>
      <rootfile full-path="OEBPS/content.opf" media-type="application/oebps-package+xml"/>
   </rootfiles>
</container>"""

    EPUB_CSS_CONTENTS = u"""@charset "UTF-8";
body {
  margin: 10px 25px 10px 25px;
}  
h1 {
  font-size: 200%;
}
h2 {
  font-size: 150%;
}
p {
  margin-left: 0em;
  margin-right: 0em;
  margin-top: 0em;
  margin-bottom: 0em;
  line-height: 2em;
  text-align: justify;
}
a, a:focus, a:active, a:visited {
  color: black;
  text-decoration: none;
}
body.indexPage {}
h1.indexTitle {}
p.indexGroups {
  font-size: 150%;
}
span.indexGroup {}
body.groupPage {}
h1.groupTitle {}
div.groupNavigation {}
span.groupHeadword {}
div.groupEntry {
  margin-top: 0;
  margin-bottom: 1em;
}
h2.groupHeadword {
  margin-left: 5%;
}
p.groupDefinition {
  margin-left: 10%;
  margin-right: 10%;
}
"""

    MOBI_CSS_CONTENTS = u""""@charset "UTF-8";"""

    INDEX_XHTML_TEMPLATE = u"""<?xml version="1.0" encoding="utf-8" standalone="no"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
 <head>
  <title>%s</title>
  <link rel="stylesheet" type="text/css" href="style.css" />
 </head>
 <body class="indexPage">
  <h1 class="indexTitle">%s</h1>
  <p class="indexGroupss">
%s
  </p>
 </body>
</html>"""
    INDEX_XHTML_LINK_TEMPLATE = u"""   <span class="indexGroup"><a href=\"%s\">%s</a></span>"""

    INDEX_XHTML_LINK_JOINER = u" &#8226;\n"

    EPUB_GROUP_XHTML_TEMPLATE = u"""<?xml version="1.0" encoding="utf-8" standalone="no"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
 <head>
  <title>%s</title>
  <link rel="stylesheet" type="text/css" href="style.css" />
 </head>
 <body id="groupPage" class="groupPage">
  <h1 class="groupTitle">%s</h1>
  <div class="groupNavigation">
   <a href="%s">[ Previous ]</a>
%s
   <a href="%s">[ Next ]</a>
  </div>
%s
 </body>
</html>"""
    EPUB_GROUP_XHTML_INDEX_LINK = u"""   <a href="index.xhtml">[ Index ]</a>"""

    EPUB_GROUP_XHTML_WORD_TEMPLATE = u"""   <span class="groupHeadword">%s</span>"""

    EPUB_GROUP_XHTML_WORD_JOINER = u" &#8226;\n"

    EPUB_GROUP_XHTML_WORD_DEFINITION_TEMPLATE = u"""  <div class="groupEntry">
   <h2 class="groupHeadword">%s</h2>
   <p class="groupDefinition">%s</p>
  </div>"""

    EPUB_GROUP_XHTML_WORD_DEFINITION_JOINER = u"\n"

    MOBI_GROUP_XHTML_TEMPLATE = u"""<?xml version="1.0" encoding="utf-8" standalone="no"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
 <head>
  <title>%s</title>
  <link rel="stylesheet" type="text/css" href="style.css" />
 </head>
 <body id="groupPage" class="groupPage">
  <h1 class="groupTitle">%s</h1>
  <div class="groupNavigation">
   <a href="%s">[ Previous ]</a>
%s
   <a href="%s">[ Next ]</a>
  </div>
%s
 </body>
</html>"""

    MOBI_GROUP_XHTML_INDEX_LINK = u"""   <a href="index.xhtml">[ Index ]</a>"""

    MOBI_GROUP_XHTML_WORD_TEMPLATE = u"""   <span class="groupHeadword"><idx:entry><idx:orth>%s</idx:orth></idx:entry></span>"""

    MOBI_GROUP_XHTML_WORD_JOINER = u" &#8226;\n"

    MOBI_GROUP_XHTML_WORD_DEFINITION_TEMPLATE = u"""  <div class="groupEntry">
   <idx:entry>
    <h2 class="groupHeadword"><idx:orth>%s</idx:orth></h2>
    <p class="groupDefinition">%s</p>
   </idx:entry>
  </div>"""

    MOBI_GROUP_XHTML_WORD_DEFINITION_JOINER = u"\n"

    EPUB2_OPF_TEMPLATE = u"""<?xml version="1.0" encoding="utf-8" ?>
<package xmlns="http://www.idpf.org/2007/opf" version="2.0" unique-identifier="uid">
 <metadata xmlns:opf="http://www.idpf.org/2007/opf" xmlns:dc="http://purl.org/dc/elements/1.1/">
  <dc:identifier id="uid" opf:scheme="uuid">%s</dc:identifier>
  <dc:language>%s</dc:language>
  <dc:title>%s</dc:title>
  <dc:creator opf:role="aut">%s</dc:creator>
  <dc:rights>%s</dc:rights>
  <dc:date opf:event="creation">%s-01-01</dc:date>
%s
 </metadata>
 <manifest>
%s
 </manifest>
 <spine toc="toc.ncx">
%s
 </spine>
</package>"""

    MOBI_OPF_TEMPLATE = u"""<?xml version="1.0" encoding="utf-8"?>
<package unique-identifier="uid">
 <metadata>
  <dc-metadata xmlns:dc="http://purl.org/metadata/dublin_core" xmlns:oebpackage="http://openebook.org/namespaces/oeb-package/1.0/">
   <dc:Title>%s</dc:Title>
   <dc:Language>%s</dc:Language>
   <dc:Identifier id="uid">%s</dc:Identifier>
   <dc:Creator>%s</dc:Creator>
   <dc:Rights>%s</dc:Rights>
   <dc:Subject BASICCode="REF008000">Dictionaries</dc:Subject>
  </dc-metadata>
  <x-metadata>
   <output encoding="utf-8"></output>
   <DictionaryInLanguage>%s</DictionaryInLanguage>
   <DictionaryOutLanguage>%s</DictionaryOutLanguage>
   <EmbeddedCover>%s</EmbeddedCover>
  </x-metadata>
 </metadata>
 <manifest>
%s
 </manifest>
 <spine>
%s
 </spine>
 <tours></tours>
 <guide></guide>
</package>"""

    OPF_MANIFEST_ITEM_TEMPLATE = u"""  <item href="%s" id="%s" media-type="%s" />"""

    OPF_SPINE_ITEMREF_TEMPLATE = u"""  <itemref idref="%s" />"""

    NCX_TEMPLATE = u"""<?xml version="1.0" encoding="utf-8" ?>
<!DOCTYPE ncx PUBLIC "-//NISO//DTD ncx 2005-1//EN" "http://www.daisy.org/z3986/2005/ncx-2005-1.dtd">
<ncx xmlns="http://www.daisy.org/z3986/2005/ncx/" version="2005-1">
 <head>
  <meta name="dtb:uid" content="%s" />
  <meta name="dtb:depth" content="1" />
  <meta name="dtb:totalPageCount" content="0" />
  <meta name="dtb:maxPageNumber" content="0" />
 </head>
 <docTitle>
  <text>%s</text>
 </docTitle>
 <navMap>
%s
 </navMap>
</ncx>"""

    NCX_NAVPOINT_TEMPLATE = u""" <navPoint id="n%06d" playOrder="%d">
  <navLabel>
   <text>%s</text>
  </navLabel>
  <content src="%s" />
 </navPoint>"""

    def __init__(self, ebook_format, args):
        self.ebook_format = ebook_format
        self.args = args
        self.root_directory_path = None
        self.cover = None
        self.files = []
        self.manifest_files = []
        self.groups = []

    def get_tmp_path(self):
        if self.root_directory_path is not None:
            return self.root_directory_path
        return u""

    def delete(self):
        if self.root_directory_path is not None:
            delete_directory(self.root_directory_path)

    def add_file(self, relative_path, contents, mode=zipfile.ZIP_DEFLATED):
        file_path = os.path.join(self.root_directory_path, relative_path)
        file_obj = open(file_path, "wb")
        try:
            # Python 2
            if isinstance(contents, unicode):
                contents = contents.encode("utf-8")
        except NameError:
            # Python 3
            if isinstance(contents, str):
                contents = contents.encode("utf-8")
        except:
            # should not occur
            pass
        file_obj.write(contents)
        file_obj.close()
        self.files.append({"path": relative_path, "mode": mode})

    def write_cover(self, cover_path_absolute):
        if cover_path_absolute is not None:
            try:
                basename = os.path.basename(cover_path_absolute)
                cover_obj = open(cover_path_absolute, "rb")
                cover = cover_obj.read()
                cover_obj.close()
                b = basename.lower()
                mimetype = "image/jpeg"
                if b.endswith(".png"):
                    mimetype = "image/png"
                elif b.endswith(".gif"):
                    mimetype = "image/gif"
                self.add_file_manifest(u"OEBPS/%s" % basename, basename, cover, mimetype)
                self.cover = basename
            except:
                pass

    def write_css(self, custom_css_path_absolute):
        if self.ebook_format == self.MOBI:
            css = self.MOBI_CSS_CONTENTS
        else:
            css = self.EPUB_CSS_CONTENTS
        if custom_css_path_absolute is not None:
            try:
                css_obj = open(custom_css_path_absolute, "rb")
                css = css_obj.read()
                css_obj.close()
            except:
                pass
        self.add_file_manifest(u"OEBPS/style.css", u"style.css", css, "text/css")

    def add_file_manifest(self, relative_path, id, contents, mimetype):
        self.add_file(relative_path, contents)
        self.manifest_files.append({"path": relative_path, "id": id, "mimetype": mimetype})

    def get_group_xhtml_file_name_from_index(self, index):
        if (index < self.GROUP_START_INDEX) or (index >= len(self.groups) + self.GROUP_START_INDEX):
            return u"#groupPage"
        return u"g%06d.xhtml" % index

    def add_group(self, key, entries):
        self.groups.append({"key": key, "entries": entries})

    def write_groups(self):
        if self.ebook_format == self.MOBI:
            group_template = self.MOBI_GROUP_XHTML_TEMPLATE
            if self.args.include_index_page:
                index_link = self.MOBI_GROUP_XHTML_INDEX_LINK
            else:
                index_link = u""
            word_template = self.MOBI_GROUP_XHTML_WORD_TEMPLATE
            word_joiner = self.MOBI_GROUP_XHTML_WORD_JOINER
            word_definition_template = self.MOBI_GROUP_XHTML_WORD_DEFINITION_TEMPLATE
            word_definition_joiner = self.MOBI_GROUP_XHTML_WORD_DEFINITION_JOINER
        else:
            group_template = self.EPUB_GROUP_XHTML_TEMPLATE
            if self.args.include_index_page:
                index_link = self.EPUB_GROUP_XHTML_INDEX_LINK
            else:
                index_link = u""
            word_template = self.EPUB_GROUP_XHTML_WORD_TEMPLATE
            word_joiner = self.EPUB_GROUP_XHTML_WORD_JOINER
            word_definition_template = self.EPUB_GROUP_XHTML_WORD_DEFINITION_TEMPLATE
            word_definition_joiner = self.EPUB_GROUP_XHTML_WORD_DEFINITION_JOINER

        index = self.GROUP_START_INDEX
        for group in self.groups:
            group_label = self.get_group_label(group)
            group_xhtml_path = self.get_group_xhtml_file_name_from_index(index)
            previous_link = self.get_group_xhtml_file_name_from_index(index - 1)
            next_link = self.get_group_xhtml_file_name_from_index(index + 1)
            group_contents = []
            if self.args.no_definitions:
                for entry in group["entries"]:
                    headword = self.escape_if_needed(entry.headword)
                    group_contents.append(word_template % (headword))
                group_contents = word_joiner.join(group_contents)
            else:
                for entry in group["entries"]:
                    headword = self.escape_if_needed(entry.headword)
                    definition = self.escape_if_needed(entry.definition)
                    group_contents.append(word_definition_template % (headword, definition))
                group_contents = word_definition_joiner.join(group_contents)
            group_contents = group_template % (group_label, group_label, previous_link, index_link, next_link, group_contents)
            self.add_file_manifest(u"OEBPS/%s" % group_xhtml_path, group_xhtml_path, group_contents, u"application/xhtml+xml")
            index += 1 

    def escape_if_needed(self, string):
        def html_escape(s):
            x = s
            x = x.replace("&", "&amp;")
            x = x.replace('"', "&quot;")
            x = x.replace("'", "&apos;")
            x = x.replace(">", "&gt;")
            x = x.replace("<", "&lt;")
            return x
        if self.args.escape_strings:
            return html_escape(string)
        return string

    def get_group_label(self, group):
        group_label = group["key"]
        if group_label != u"SPECIAL":
            group_label = "%s&#8211;%s" % (group["entries"][0].headword, group["entries"][-1].headword)
        return group_label

    def write_index(self):
        links = []
        index = self.GROUP_START_INDEX
        for group in self.groups:
            group_label = self.get_group_label(group)
            group_xhtml_path = self.get_group_xhtml_file_name_from_index(index)
            group_link = self.INDEX_XHTML_LINK_TEMPLATE % (group_xhtml_path, group_label)
            links.append(group_link)
            index += 1
        links = self.INDEX_XHTML_LINK_JOINER.join(links)
        contents = self.INDEX_XHTML_TEMPLATE % (self.args.title, self.args.title, links)
        self.add_file_manifest(u"OEBPS/index.xhtml", u"index.xhtml", contents, u"application/xhtml+xml")

    def write_opf(self):
        manifest_contents = []
        spine_contents = []
        for mi in self.manifest_files:
            manifest_contents.append(self.OPF_MANIFEST_ITEM_TEMPLATE % (mi["id"], mi["id"], mi["mimetype"]))
            if mi["mimetype"] == u"application/xhtml+xml":
                spine_contents.append(self.OPF_SPINE_ITEMREF_TEMPLATE % (mi["id"]))
        manifest_contents = u"\n".join(manifest_contents)
        spine_contents = u"\n".join(spine_contents)
        cover = u""
        if self.ebook_format == self.MOBI:
            if self.cover is not None:
                cover = self.cover
            opf_contents = self.MOBI_OPF_TEMPLATE % (
                self.args.title,
                self.args.language_from,
                self.args.identifier,
                self.args.author,
                self.args.copyright,
                self.args.language_from,
                self.args.language_to,
                cover,
                manifest_contents,
                spine_contents
            )
        else:
            if self.cover is not None:
                cover = u"""  <meta name="cover" content="%s" />""" % self.cover
            opf_contents = self.EPUB2_OPF_TEMPLATE % (
                self.args.identifier,
                self.args.language_from,
                self.args.title,
                self.args.author,
                self.args.copyright,
                self.args.year,
                cover,
                manifest_contents,
                spine_contents
            )
        self.add_file("OEBPS/content.opf", opf_contents)

    def write_ncx(self):
        ncx_items = []
        index = 1
        if self.args.include_index_page:
            ncx_items.append(self.NCX_NAVPOINT_TEMPLATE % (index, index, "Index", "index.xhtml"))
            index += 1
        for group in self.groups:
            group_label = self.get_group_label(group)
            group_xhtml_path = self.get_group_xhtml_file_name_from_index(index)
            ncx_items.append(self.NCX_NAVPOINT_TEMPLATE % (index, index, group_label, group_xhtml_path))
            index += 1
        ncx_items = u"\n".join(ncx_items)
        ncx_contents = self.NCX_TEMPLATE % (self.args.identifier, self.args.title, ncx_items)
        self.add_file_manifest(u"OEBPS/toc.ncx", u"toc.ncx", ncx_contents, u"application/x-dtbncx+xml")

    def write(self, file_path_absolute, compress=True): 
        # get cover path
        cover_path_absolute = self.args.cover_path
        if cover_path_absolute is not None:
            cover_path_absolute = os.path.abspath(cover_path_absolute)

        # get custom css path
        custom_css_path_absolute = self.args.apply_css
        if custom_css_path_absolute is not None:
            custom_css_path_absolute = os.path.abspath(custom_css_path_absolute)

        # create new tmp directory and cd there
        self.root_directory_path = create_temp_directory()
        cwd = os.getcwd()
        os.chdir(self.root_directory_path)
        os.makedirs(u"META-INF")
        os.makedirs(u"OEBPS")

        # add mimetype and container.xml
        if self.ebook_format in [self.EPUB2]: # add EPUB3 here
            self.add_file(u"mimetype", self.MIMETYPE_CONTENTS, mode=zipfile.ZIP_STORED)
            self.add_file(u"META-INF/container.xml", self.CONTAINER_XML_CONTENTS)

        # add cover
        self.write_cover(cover_path_absolute)

        # write CSS
        self.write_css(custom_css_path_absolute)

        # write index
        if self.args.include_index_page:
            self.write_index()

        # write groups
        self.write_groups()

        # write ncx
        if self.ebook_format in [self.EPUB2]: # add EPUB3 here
            self.write_ncx()

        # write opf
        self.write_opf()

        # compress 
        if compress:
            output_file_obj = zipfile.ZipFile(file_path_absolute, "w", compression=zipfile.ZIP_DEFLATED)
            for file_to_compress in self.files:
                output_file_obj.write(file_to_compress["path"], compress_type=file_to_compress["mode"])
            output_file_obj.close()

        # return to previous cwd
        os.chdir(cwd)





#!/usr/bin/env python
# coding=utf-8

"""
This file contains command line constants and functions.
"""

from __future__ import absolute_import
import datetime
import sys

from penelope.utilities import get_uuid
from penelope.utilities import print_error

__author__ = "Alberto Pettarin"
__copyright__ = "Copyright 2012-2015, Alberto Pettarin (www.albertopettarin.it)"
__license__ = "MIT"
__version__ = "3.1.0"
__email__ = "alberto@albertopettarin.it"
__status__ = "Production"

INPUT_FORMATS = [
    "bookeen",
    "csv",
    "kobo",
    "stardict",
    "xml"
]

OUTPUT_FORMATS = [
    "bookeen",
    "csv",
    "epub",
    "kobo",
    "mobi",
    "stardict",
    "xml"
]

COMMAND_LINE_PARAMETERS = [
    {
        "short": "-d",
        "long": "--debug",
        "help": "enable debug mode (default: False)",
        "action": "store_true"
    },
    {
        "short": "-f",
        "long": "--language-from",
        "help": "from language (ISO 639-1 code)",
        "action": "store"
    },
    {
        "short": "-i",
        "long": "--input-file",
        "help": "input file name prefix(es). Multiple prefixes must be comma-separated.",
        "action": "store"
    },
    {
        "short": "-j",
        "long": "--input-format",
        "help": "from format (values: %s)" % "|".join(INPUT_FORMATS),
        "action": "store"
    },
    {
        "short": "-k",
        "long": "--keep",
        "help": "keep temporary files (default: False)",
        "action": "store_true"
    },
    {
        "short": "-o",
        "long": "--output-file",
        "help": "output file name",
        "action": "store"
    },
    {
        "short": "-p",
        "long": "--output-format",
        "help": "to format (values: %s)" % "|".join(OUTPUT_FORMATS),
        "action": "store"
    },
    {
        "short": "-t",
        "long": "--language-to",
        "help": "to language (ISO 639-1 code)",
        "action": "store"
    },
    {
        "short": "-v",
        "long": "--version",
        "help": "print version and exit",
        "action": "store_true"
    },

    {
        "short": None,
        "long": "--author",
        "help": "author string",
        "action": "store"
    },
    {
        "short": None,
        "long": "--copyright",
        "help": "copyright string",
        "action": "store"
    },
    {
        "short": None,
        "long": "--cover-path",
        "help": "path of the cover image file",
        "action": "store"
    },
    {
        "short": None,
        "long": "--description",
        "help": "description string",
        "action": "store"
    },
    {
        "short": None,
        "long": "--email",
        "help": "email string",
        "action": "store"
    },
    {
        "short": None,
        "long": "--identifier",
        "help": "identifier string",
        "action": "store"
    },
    {
        "short": None,
        "long": "--license",
        "help": "license string",
        "action": "store"
    },
    {
        "short": None,
        "long": "--title",
        "help": "title string",
        "action": "store"
    },
    {
        "short": None,
        "long": "--website",
        "help": "website string",
        "action": "store"
    },
    {
        "short": None,
        "long": "--year",
        "help": "year string",
        "action": "store"
    },

    {
        "short": None,
        "long": "--apply-css",
        "help": "apply the given CSS file (epub and mobi output only)",
        "action": "store"
    },
    {
        "short": None,
        "long": "--bookeen-collation-function",
        "help": "use the specified collation function",
        "action": "store"
    },
    {
        "short": None,
        "long": "--bookeen-install-file",
        "help": "create *.install file (default: False)",
        "action": "store_true"
    },
    {
        "short": None,
        "long": "--csv-fs",
        "help": "CSV field separator (default: ',')",
        "action": "store"
    },
    {
        "short": None,
        "long": "--csv-ignore-first-line",
        "help": "ignore the first line of the input CSV file(s) (default: False)",
        "action": "store_true"
    },
    {
        "short": None,
        "long": "--csv-ls",
        "help": "CSV line separator (default: '\\n')",
        "action": "store"
    },
    {
        "short": None,
        "long": "--dictzip-path",
        "help": "path to dictzip executable",
        "action": "store"
    },
    {
        "short": None,
        "long": "--epub-no-compress",
        "help": "do not create the compressed container (epub output only, default: False)",
        "action": "store_true"
    },
    {
        "short": None,
        "long": "--escape-strings",
        "help": "escape HTML strings (default: False)",
        "action": "store_true"
    },
    {
        "short": None,
        "long": "--flatten-synonyms",
        "help": "flatten synonyms, creating a new entry with headword=synonym and using the definition of the original headword (default: False)",
        "action": "store_true"
    },
    {
        "short": None,
        "long": "--group-by-prefix-function",
        "help": "compute the prefix of headwords using the given prefix function file",
        "action": "store"
    },
    {
        "short": None,
        "long": "--group-by-prefix-length",
        "help": "group headwords by prefix of given length (default: 2)",
        "action": "store"
    },
    {
        "short": None,
        "long": "--group-by-prefix-merge-across-first",
        "help": "merge headword groups even when the first character changes (default: False)",
        "action": "store_true"
    },
    {
        "short": None,
        "long": "--group-by-prefix-merge-min-size",
        "help": "merge headword groups until the given minimum number of headwords is reached (default: 0, meaning no merge will take place)",
        "action": "store"
    },
    {
        "short": None,
        "long": "--ignore-case",
        "help": "ignore headword case, all headwords will be lowercased (default: False)",
        "action": "store_true"
    },
    {
        "short": None,
        "long": "--ignore-synonyms",
        "help": "ignore synonyms, not reading/writing them if present (default: False)",
        "action": "store_true"
    },
    {
        "short": None,
        "long": "--include-index-page",
        "help": "include an index page (epub and mobi output only, default: False)",
        "action": "store_true"
    },
    {
        "short": None,
        "long": "--input-file-encoding",
        "help": "use the specified encoding for reading the raw contents of input file(s) (default: 'utf-8')",
        "action": "store"
    },
    {
        "short": None,
        "long": "--input-parser",
        "help": "use the specified parser function after reading the raw contents of input file(s)",
        "action": "store"
    },
    {
        "short": None,
        "long": "--kindlegen-path",
        "help": "path to kindlegen executable",
        "action": "store"
    },
    {
        "short": None,
        "long": "--marisa-bin-path",
        "help": "path to MARISA bin directory",
        "action": "store"
    },
    {
        "short": None,
        "long": "--marisa-index-size",
        "help": "maximum size of the MARISA index (default: 1000000)",
        "action": "store"
    },
    {
        "short": None,
        "long": "--merge-definitions",
        "help": "merge definitions for the same headword (default: False)",
        "action": "store_true"
    },
    {
        "short": None,
        "long": "--merge-separator",
        "help": "add this string between merged definitions (default: ' | ')",
        "action": "store"
    },
    {
        "short": None,
        "long": "--mobi-no-kindlegen",
        "help": "do not run kindlegen, keep .opf and .html files (default: False)",
        "action": "store_true"
    },
    {
        "short": None,
        "long": "--no-definitions",
        "help": "do not output definitions for EPUB and MOBI formats (default: False)",
        "action": "store_true"
    },
    {
        "short": None,
        "long": "--sd-ignore-sametypesequence",
        "help": "ignore the value of sametypesequence in StarDict .ifo files (default: False)",
        "action": "store_true"
    },
    {
        "short": None,
        "long": "--sd-no-dictzip",
        "help": "do not compress the .dict file in StarDict files (default: False)",
        "action": "store_true"
    },
    {
        "short": None,
        "long": "--sort-after",
        "help": "sort after merging/flattening (default: False)",
        "action": "store_true"
    },
    {
        "short": None,
        "long": "--sort-before",
        "help": "sort before merging/flattening (default: False)",
        "action": "store_true"
    },
    {
        "short": None,
        "long": "--sort-by-definition",
        "help": "sort by definition (default: False)",
        "action": "store_true"
    },
    {
        "short": None,
        "long": "--sort-by-headword",
        "help": "sort by headword (default: False)",
        "action": "store_true"
    },
    {
        "short": None,
        "long": "--sort-ignore-case",
        "help": "ignore case when sorting (default: False)",
        "action": "store_true"
    },
    {
        "short": None,
        "long": "--sort-reverse",
        "help": "reverse the sort order (default: False)",
        "action": "store_true"
    },
]

REQUIRED_PARAMETERS = [
    "input_file",
    "input_format",
    "language_from",
    "language_to",
    "output_format",
    "output_file"
]

EXAMPLES = [
    #{
    #    "options": "-h",
    #    "description": "Print this message and exit"
    #},
    #{
    #    "options": "-v",
    #    "description": "Print the version and exit"
    #},
    {
        "options": "-i dict.csv -j csv -f en -t it -p stardict -o output.zip",
        "description": "Convert en->it dictionary dict.csv (in CSV format) into output.zip (in StarDict format)"
    },
    {
        "options": "-i dict.csv -j csv -f en -t it -p stardict -o output.zip --merge-definitions",
        "description": "As above, but also merge definitions"
    },
    {
        "options": "-i d1,d2,d3 -j csv -f en -t it -p csv -o output.csv --sort-after --sort-by-headword",
        "description": "Merge CSV dictionaries d1, d2, and d3 into output.csv, sorting by headword"
    },
    {
        "options": "-i d1,d2,d3 -j csv -f en -t it -p csv -o output.csv --sort-after --sort-by-headword --sort-ignore-case",
        "description": "As above, but ignore case for sorting"
    },
    {
        "options": "-i d1,d2,d3 -j csv -f en -t it -p csv -o output.csv --sort-after --sort-by-headword --sort-reverse",
        "description": "As above, but reverse the order"
    },
    {
        "options": "-i dict.zip -j stardict -f en -t it -p csv -o output.csv",
        "description": "Convert en->it dictionary dict.zip (in StarDict format) into output.csv (in CSV format)"
    },
    {
        "options": "-i dict.zip -j stardict -f en -t it -p csv -o output.csv --ignore-synonyms",
        "description": "As above, but do not read the .syn synonym file if present"
    },
    {
        "options": "-i dict.zip -j stardict -f en -t it -p csv -o output.csv --flatten-synonyms",
        "description": "As above, but flatten synonyms"
    },
    {
        "options": "-i dict.zip -j stardict -f en -t it -p bookeen -o output",
        "description": "Convert dict.zip into output.dict.idx and output.dict for Bookeen devices"
    },
    {
        "options": "-i dict.zip -j stardict -f en -t it -p kobo -o dicthtml-en-it",
        "description": "Convert dict.zip into dicthtml-en-it.zip for Kobo devices"
    },
    {
        "options": "-i dict.csv -j csv -f en -t it -p mobi -o output.mobi --cover-path mycover.png --title \"My English->Italian Dictionary\"",
        "description": "Convert dict.csv into a MOBI (Kindle) dictionary, using the specified cover image and title"
    },
    {
        "options": "-i dict.xml -j xml -f en -t it -p mobi -o output.epub",
        "description": "Convert dict.xml into an EPUB dictionary"
    },
    {
        "options": "-i dict.xml -j xml -f en -t it -p mobi -o output.epub --epub-output-definitions",
        "description": "As above, but also output definitions"
    },
]

USAGE = u"""
  $ penelope -h
  $ penelope -i INPUT_FILE -j INPUT_FORMAT -f LANGUAGE_FROM -t LANGUAGE_TO -p OUTPUT_FORMAT -o OUTPUT_FILE [OPTIONS]
  $ penelope -i IN1,IN2[,IN3...] -j INPUT_FORMAT -f LANGUAGE_FROM -t LANGUAGE_TO -p OUTPUT_FORMAT -o OUTPUT_FILE [OPTIONS]
"""

DESCRIPTION = u"""description:
  Convert dictionary file(s) with file name prefix INPUT_FILE from format INPUT_FORMAT to format OUTPUT_FORMAT, saving it as OUTPUT_FILE.
  The dictionary is from LANGUAGE_FROM to LANGUAGE_TO, possibly the same.
  You can merge several dictionaries (with the same format), by providing a list of comma-separated prefixes, as shown by the third synopsis above."""

EPILOG = u"examples:\n"
for example in EXAMPLES:
    EPILOG += u"\n"
    EPILOG += u"  $ penelope %s\n" % (example["options"])
    EPILOG += u"    %s\n" % (example["description"])
EPILOG += u"  \n"

def check_arguments(args):
    """
    Check that we have all the required command line arguments,
    and that the input/output format values are supported.
    """
    for required in REQUIRED_PARAMETERS:
        if required not in args:
            print_error("Argument '%s' is required" % required)
            sys.exit(2)
    if args.input_format not in INPUT_FORMATS:
        print_error("Format '%s' is not a valid input format" % args.input_format)
        print_error("Valid input formats: %s" % INPUT_FORMATS)
        sys.exit(4)
    if args.output_format not in OUTPUT_FORMATS:
        print_error("Format '%s' is not a valid output format" % args.output_format)
        print_error("Valid output formats: %s" % OUTPUT_FORMATS)
        sys.exit(4)

def set_default_values(args):
    def set_default_value(key, value):
        if not args.__contains__(key):
            args.__dict__[key] = value
    set_default_value("apply_css", None)
    set_default_value("bookeen_collation_function", None)
    set_default_value("bookeen_install_file", False)
    set_default_value("csv_fs", ",")
    set_default_value("csv_ignore_first_line", False)
    set_default_value("csv_ls", "\n")
    set_default_value("debug", False)
    set_default_value("dictzip_path", None)
    set_default_value("epub_no_compress", False)
    set_default_value("escape_strings", False)
    set_default_value("flatten_synonyms", False)
    set_default_value("group_by_prefix_length", 2)
    set_default_value("group_by_prefix_function", None)
    set_default_value("group_by_prefix_merge_across_first", False)
    set_default_value("group_by_prefix_merge_min_size", 0)
    set_default_value("ignore_case", False)
    set_default_value("ignore_synonyms", False)
    set_default_value("include_index_page", False)
    set_default_value("input_file_encoding", "utf-8")
    set_default_value("input_parser", None)
    set_default_value("keep", False)
    set_default_value("kindlegen_path", None)
    set_default_value("marisa_bin_path", None)
    set_default_value("marisa_index_size", 1000000)
    set_default_value("merge_definitions", False)
    set_default_value("merge_separator", " | ")
    set_default_value("mobi_no_kindlegen", False)
    set_default_value("no_definitions", False)
    set_default_value("sd_ignore_sametypesequence", False)
    set_default_value("sd_no_dictzip", False)
    set_default_value("sort_after", False)
    set_default_value("sort_before", False)
    set_default_value("sort_by_definition", False)
    set_default_value("sort_by_headword", False)
    set_default_value("sort_ignore_case", False)
    set_default_value("sort_reverse", False)
    set_default_value("version", False)
    set_default_value("author", u"Penelope")
    set_default_value("copyright", u"GNU GPL v3")
    set_default_value("cover_path", None)
    set_default_value("description", u"Dictionary %s to %s" % (args.language_from, args.language_to))
    set_default_value("email", u"penelopedictionaryconverter@gmail.com")
    set_default_value("identifier", get_uuid())
    set_default_value("license", u"GNU GPL v3")
    set_default_value("title", u"Dictionary %s to %s" % (args.language_from, args.language_to))
    set_default_value("website", u"https://goo.gl/EB5XSR")
    set_default_value("year", str(datetime.datetime.now().year))



